# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import time
import typing

import netaddr
from oslo_log import log

import tobiko
from tobiko.shell import sh
from tobiko.shell.ping import _interface
from tobiko.shell.ping import _exception
from tobiko.shell.ping import _parameters
from tobiko.shell.ping import _statistics


LOG = log.getLogger(__name__)


TRANSMITTED = 'transmitted'
DELIVERED = 'delivered'
UNDELIVERED = 'undelivered'
RECEIVED = 'received'
UNRECEIVED = 'unreceived'


PingHostType = typing.Union['str', netaddr.IPAddress]


def list_reachable_hosts(hosts: typing.Iterable[PingHostType],
                         **params) -> tobiko.Selection[PingHostType]:
    reachable_host, _ = ping_hosts(hosts, **params)
    return reachable_host


def list_unreachable_hosts(hosts: typing.Iterable[PingHostType],
                           **params) -> tobiko.Selection[PingHostType]:
    _, unreachable_host = ping_hosts(hosts, **params)
    return unreachable_host


PingHostsResultType = typing.Tuple[tobiko.Selection[PingHostType],
                                   tobiko.Selection[PingHostType]]


def ping_hosts(hosts: typing.Iterable[PingHostType],
               count: typing.Optional[int] = None,
               **params) -> PingHostsResultType:
    if count is None:
        count = 1
    else:
        count = int(count)
    reachable = tobiko.Selection[PingHostType]()
    unreachable = tobiko.Selection[PingHostType]()
    for host in hosts:
        try:
            result = ping(host, count=count, **params)
        except _exception.PingError:
            LOG.exception('Error pinging host: %r', host)
            unreachable.append(host)
        else:
            if result.received:
                reachable.append(host)
            else:
                unreachable.append(host)
    return reachable, unreachable


def ping(host: PingHostType, until=TRANSMITTED, check: bool = True,
         **ping_params) -> _statistics.PingStatistics:
    """Send ICMP messages to host address until timeout

    :param host: destination host address
    :param **ping_params: parameters to be forwarded to get_statistics()
        function
    :returns: PingStatistics
    """
    return get_statistics(host=host, until=until, check=check, **ping_params)


def ping_until_delivered(host, **ping_params):
    """Send 'count' ICMP messages

    Send 'count' ICMP messages

    ICMP messages are considered delivered when they have been
    transmitted without being counted as errors.

    :param host: destination host address
    :param **ping_params: parameters to be forwarded to get_statistics()
        function
    :returns: PingStatistics
    :raises: PingFailed in case timeout expires before delivering all
        expected count messages
    """
    return ping(host=host, until=DELIVERED, **ping_params)


def ping_until_undelivered(host, **ping_params):
    """Send ICMP messages until it fails to deliver messages

    Send ICMP messages until it fails to deliver 'count' messages

    ICMP messages are considered undelivered when they have been
    transmitted and they have been counted as error in ping statistics (for
    example because of errors into the route to remote address).

    :param host: destination host address
    :param **ping_params: parameters to be forwarded to get_statistics()
        function
    :returns: PingStatistics
    :raises: PingFailed in case timeout expires before failing delivering
        expected 'count' of messages
    """
    return ping(host=host, until=UNDELIVERED, **ping_params)


def ping_until_received(host, **ping_params):
    """Send ICMP messages until it receives messages back

    Send ICMP messages until it receives 'count' messages back

    ICMP messages are considered received when they have been
    transmitted without any routing errors and they are received back

    :param host: destination host address
    :param **ping_params: parameters to be forwarded to get_statistics()
        function
    :returns: PingStatistics
    :raises: PingFailed in case timeout expires before receiving all
        expected 'count' of messages
    """
    return ping(host=host, until=RECEIVED, **ping_params)


def ping_until_unreceived(host, **ping_params):
    """Send ICMP messages until it fails to receive messages

    Send ICMP messages until it fails to receive 'count' messages back.

    ICMP messages are considered unreceived when they have been
    transmitted without any routing error but they failed to be received
    back (for example because of network filtering).

    :param host: destination host address
    :param **ping_params: parameters to be forwarded to get_statistics()
        function
    :returns: PingStatistics
    :raises: PingFailed in case timeout expires before failed receiving
        expected 'count' of messages
    """
    return ping(host=host, until=UNRECEIVED, **ping_params)


def get_statistics(parameters=None, ssh_client=None, until=None, check=True,
                   **ping_params) -> _statistics.PingStatistics:
    parameters = _parameters.get_ping_parameters(default=parameters,
                                                 **ping_params)
    statistics = _statistics.PingStatistics()
    for partial_statistics in iter_statistics(parameters=parameters,
                                              ssh_client=ssh_client,
                                              until=until, check=check):
        statistics += partial_statistics
        LOG.debug("%r", statistics)

    return statistics


def iter_statistics(parameters=None, ssh_client=None, until=None, check=True,
                    **ping_params):
    parameters = _parameters.get_ping_parameters(default=parameters,
                                                 **ping_params)
    now = time.time()
    end_of_time = now + parameters.timeout
    deadline = parameters.deadline
    transmitted = 0
    received = 0
    undelivered = 0
    count = 0
    enlapsed_time = None

    while deadline > 0. and count < parameters.count:
        if enlapsed_time is not None and enlapsed_time < deadline:
            # Avoid busy waiting when errors happens
            sleep_time = deadline - enlapsed_time
            LOG.debug('Waiting %s seconds before next ping execution',
                      sleep_time)
            time.sleep(sleep_time)

        start_time = time.time()

        # splitting total timeout interval into smaller deadline intervals will
        # cause ping command to be executed more times allowing to handle
        # temporary packets routing problems
        if until == RECEIVED:
            execute_parameters = _parameters.get_ping_parameters(
                default=parameters,
                deadline=deadline,
                count=(parameters.count - count),
                timeout=end_of_time - now)
        else:
            # Deadline ping parameter cause ping to be executed until count
            # messages are received or deadline is expired
            # Therefore to count messages not of received type we have to
            # simulate deadline parameter limiting the maximum number of
            # transmitted messages
            execute_parameters = _parameters.get_ping_parameters(
                default=parameters,
                deadline=deadline,
                count=min(parameters.count - count,
                          parameters.interval * deadline),
                timeout=end_of_time - now)

        # Command timeout would typically give ping command additional seconds
        # to safely reach deadline before shell command timeout expires, while
        # in the same time adding an extra verification to forbid using more
        # time than expected considering the time required to make SSH
        # connection and running a remote shell
        output = execute_ping(parameters=execute_parameters,
                              ssh_client=ssh_client,
                              check=check)

        if output:
            statistics = _statistics.parse_ping_statistics(
                output=output, begin_interval=now,
                end_interval=time.time())

            yield statistics

            transmitted += statistics.transmitted
            received += statistics.received
            undelivered += statistics.undelivered
        else:
            # Assume 1 transmitted undelivered package when unable to get
            # ping output
            transmitted += 1
            undelivered += 1

        count = {None: 0,
                 TRANSMITTED: transmitted,
                 DELIVERED: transmitted - undelivered,
                 UNDELIVERED: undelivered,
                 RECEIVED: received,
                 UNRECEIVED: transmitted - received}[until]

        now = time.time()
        deadline = min(int(end_of_time - now), parameters.deadline)
        enlapsed_time = now - start_time
        if enlapsed_time > 0.:
            LOG.debug('Ping execution took %s seconds', enlapsed_time)

    if until and count < parameters.count:
        raise _exception.PingFailed(count=count,
                                    expected_count=parameters.count,
                                    timeout=parameters.timeout,
                                    message_type=until)


def execute_ping(parameters, ssh_client=None, check=True):
    command = _interface.get_ping_command(parameters=parameters,
                                          ssh_client=ssh_client)

    try:
        result = sh.execute(command=command,
                            ssh_client=ssh_client,
                            timeout=parameters.deadline + 2.,
                            expect_exit_status=None,
                            network_namespace=parameters.network_namespace)
    except sh.ShellError as ex:
        LOG.exception("Error executing ping command")
        stdout = ex.stdout
        stderr = ex.stderr
    else:
        stdout = result.stdout
        stderr = result.stderr

    if stdout:
        output = str(stdout)
    else:
        output = None

    if stderr:
        error = str(stderr)
        if check and result.exit_status:
            handle_ping_command_error(error=error)

    return output


def handle_ping_command_error(error):
    for error in error.splitlines():
        error = error.strip()
        if error:
            prefix = 'ping: '
            if error.startswith('ping: '):
                error = error[len(prefix):]
            handle_ping_bad_address_error(error)
            handle_ping_local_error(error)
            handle_ping_connect_error(error)
            handle_ping_send_to_error(error)
            handle_ping_unknow_host_error(error)
            raise _exception.PingError(details=error)


def handle_ping_bad_address_error(text):
    prefix = 'bad address '
    if text.startswith(prefix):
        address = text[len(prefix):].replace("'", '').strip()
        raise _exception.BadAddressPingError(address=address)


def handle_ping_local_error(text):
    prefix = 'local error: '
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.LocalPingError(details=details)


def handle_ping_connect_error(text):
    prefix = 'connect: '
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.ConnectPingError(details=details)


def handle_ping_send_to_error(text):
    prefix = 'sendto: '
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.SendToPingError(details=details)


def handle_ping_unknow_host_error(text):
    prefix = 'unknown host'
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.UnknowHostError(details=details)

    prefix = 'unreachable-host: '
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.UnknowHostError(details=details)

    suffix = ': Name or service not known'
    if text.endswith(suffix):
        details = text[:-len(suffix)].strip()
        raise _exception.UnknowHostError(details=details)

    suffix = ': No route to host'
    if text.endswith(suffix):
        details = text[:-len(suffix)].strip()
        raise _exception.UnknowHostError(details=details)

    suffix = ': Unknown host'
    if text.endswith(suffix):
        details = text[:-len(suffix)].strip().split()[-1]
        raise _exception.UnknowHostError(details=details)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import time

from novaclient import client as novaclient
from novaclient.v2 import client as client_v2
from oslo_log import log

import tobiko
from tobiko.openstack import _client


CLIENT_CLASSES = (client_v2.Client,)
LOG = log.getLogger(__name__)


class NovaClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session):
        return novaclient.Client('2', session=session)


class NovaClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return NovaClientFixture(session=session)


CLIENTS = NovaClientManager()


def nova_client(obj):
    if not obj:
        return get_nova_client()

    if isinstance(obj, CLIENT_CLASSES):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, NovaClientFixture):
        return fixture.client

    message = "Object {!r} is not a NovaClientFixture".format(obj)
    raise TypeError(message)


def get_nova_client(session=None, shared=True, init_client=None,
                    manager=None):
    manager = manager or CLIENTS
    client = manager.get_client(session=session, shared=shared,
                                init_client=init_client)
    tobiko.setup_fixture(client)
    return client.client


def list_hypervisors(client=None, detailed=True, **params):
    client = nova_client(client)
    hypervisors = client.hypervisors.list(detailed=detailed)
    return tobiko.select(hypervisors).with_attributes(**params)


def find_hypervisor(client=None, unique=False, **params):
    hypervisors = list_hypervisors(client=client, **params)
    if unique:
        return hypervisors.unique
    else:
        return hypervisors.first


def list_servers(client=None, **params):
    client = nova_client(client)
    servers = client.servers.list()
    return tobiko.select(servers).with_attributes(**params)


def find_server(client=None, unique=False, **params):
    servers = list_servers(client=client, **params)
    if unique:
        return servers.unique
    else:
        return servers.first


def list_services(client=None, **params) -> tobiko.Selection:
    client = nova_client(client)
    services = client.services.list()
    return tobiko.select(services).with_attributes(**params)


def find_service(client=None, unique=False, **params):
    services = list_services(client=client, **params)
    if unique:
        return services.unique
    else:
        return services.first


def get_server_id(server):
    if isinstance(server, str):
        return server
    else:
        return server.id


def get_server(server, client=None, **params):
    server_id = get_server_id(server)
    return nova_client(client).servers.get(server_id, **params)


def migrate_server(server, client=None, **params):
    # pylint: disable=protected-access
    server_id = get_server_id(server)
    LOG.debug(f"Start server migration (server_id='{server_id}', "
              f"info={params})")
    return nova_client(client).servers._action('migrate', server_id,
                                               info=params)


def confirm_resize(server, client=None, **params):
    server_id = get_server_id(server)
    LOG.debug(f"Confirm server resize (server_id='{server_id}', "
              f"info={params})")
    return nova_client(client).servers.confirm_resize(server_id, **params)


MAX_SERVER_CONSOLE_OUTPUT_LENGTH = 1024 * 256


def get_console_output(server, timeout=None, interval=1., length=None,
                       client=None):
    client = nova_client(client)
    start_time = time.time()
    if length is not None:
        length = min(length, MAX_SERVER_CONSOLE_OUTPUT_LENGTH)
    else:
        length = MAX_SERVER_CONSOLE_OUTPUT_LENGTH
    while True:
        try:
            output = client.servers.get_console_output(server=server,
                                                       length=length)
        except TypeError:
            # For some reason it could happen resulting body cannot be
            # translated to json object and it is converted to None
            # on such case get_console_output would raise a TypeError
            return None

        if timeout is None or output:
            break

        if time.time() - start_time > timeout:
            LOG.warning("No console output produced by server (%r) after "
                        "%r seconds", server, timeout)
            break

        LOG.debug('Waiting for server (%r) console output...', server)
        time.sleep(interval)

    return output


class HasNovaClientMixin(object):

    nova_client = None

    def get_server(self, server, **params):
        return get_server(server=server, client=self.nova_client, **params)

    def get_server_console_output(self, server, **params):
        return get_console_output(server=server, client=self.nova_client,
                                  **params)


class WaitForServerStatusError(tobiko.TobikoException):
    message = ("Server {server_id} not changing status from {server_status} "
               "to {status}")


class WaitForServerStatusTimeout(WaitForServerStatusError):
    message = ("Server {server_id} didn't change its status from "
               "{server_status} to {status} status after {timeout} seconds")


NOVA_SERVER_TRANSIENT_STATUS = {
    'ACTIVE': ('BUILD', 'SHUTOFF'),
    'SHUTOFF': ('ACTIVE'),
    'VERIFY_RESIZE': ('RESIZE'),
}


def wait_for_server_status(server, status, client=None, timeout=None,
                           sleep_time=None, transient_status=None):
    if timeout is None:
        timeout = 300.
    if sleep_time is None:
        sleep_time = 5.
    start_time = time.time()
    if transient_status is None:
        transient_status = NOVA_SERVER_TRANSIENT_STATUS.get(status) or tuple()
    while True:
        server = get_server(server=server, client=client)
        if server.status == status:
            break

        if server.status not in transient_status:
            raise WaitForServerStatusError(server_id=server.id,
                                           server_status=server.status,
                                           status=status)

        if time.time() - start_time >= timeout:
            raise WaitForServerStatusTimeout(server_id=server.id,
                                             server_status=server.status,
                                             status=status,
                                             timeout=timeout)

        progress = getattr(server, 'progress', None)
        LOG.debug(f"Waiting for server {server.id} status to get from "
                  f"{server.status} to {status} "
                  f"(progress={progress}%)")
        time.sleep(sleep_time)
    return server


def shutoff_server(server, client=None, timeout=None, sleep_time=None):
    client = nova_client(client)
    server = get_server(server=server, client=client)
    if server.status == 'SHUTOFF':
        return server

    client.servers.stop(server.id)
    return wait_for_server_status(server=server.id, status='SHUTOFF',
                                  client=client, timeout=timeout,
                                  sleep_time=sleep_time)


def activate_server(server, client=None, timeout=None, sleep_time=None):
    client = nova_client(client)
    server = get_server(server=server, client=client)
    if server.status == 'ACTIVE':
        return server

    if server.status == 'SHUTOFF':
        client.servers.start(server.id)
    elif server.status == 'RESIZE':
        wait_for_server_status(server=server.id, status='VERIFY_RESIZE',
                               client=client, timeout=timeout,
                               sleep_time=sleep_time)
        client.servers.confirm_resize(server)
    elif server.status == 'VERIFY_RESIZE':
        client.servers.confirm_resize(server)
    else:
        client.servers.reboot(server.id, reboot_type='HARD')

    return wait_for_server_status(server=server.id, status='ACTIVE',
                                  client=client, timeout=timeout,
                                  sleep_time=sleep_time)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import typing

import netaddr
from oslo_log import log

import tobiko
from tobiko.openstack.topology import _config
from tobiko.shell import ssh

LOG = log.getLogger(__name__)


class UreachableSSHServer(tobiko.TobikoException):
    message = ("Unable to reach SSH server through any address: {addresses}. "
               "Failures: {failures}")


class SSHConnection(object):

    def __init__(self,
                 address: netaddr.IPAddress,
                 ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
                 proxy_client: typing.Optional[ssh.SSHClientFixture] = None,
                 failure: typing.Optional[Exception] = None):
        self.address = address
        self.ssh_client = ssh_client
        self.proxy_client = proxy_client
        self.failure = failure

    def __repr__(self) -> str:
        attributes = ", ".join(f"{n}={v!r}"
                               for n, v in self._iter_attributes())
        return f"{type(self).__name__}({attributes})"

    def _iter_attributes(self):
        yield 'address', self.address
        if self.ssh_client is not None:
            yield 'ssh_client', self.ssh_client
        if self.proxy_client is not None:
            yield 'proxy_client', self.proxy_client
        if self.failure is not None:
            yield 'failure', self.failure

    @property
    def is_valid(self) -> bool:
        return (self.failure is None and
                self.ssh_client is not None)


class SSHConnectionManager(tobiko.SharedFixture):

    config = tobiko.required_setup_fixture(_config.OpenStackTopologyConfig)

    def __init__(self):
        super(SSHConnectionManager, self).__init__()
        self._connections: typing.Dict[netaddr.IPAddress, SSHConnection] = (
            collections.OrderedDict())

    def cleanup_fixture(self):
        connections = list(self._connections.values())
        self._connections.clear()
        for connection in connections:
            connection.close()

    def connect(self,
                addresses: typing.List[netaddr.IPAddress],
                proxy_client: typing.Optional[ssh.SSHClientFixture] = None,
                **connect_parameters) -> ssh.SSHClientFixture:
        if not addresses:
            raise ValueError(f"'addresses' list is empty: {addresses}")

        connections = tobiko.select(
            self.list_connections(addresses, proxy_client=proxy_client))
        try:
            return connections.with_attributes(is_valid=True).first.ssh_client
        except tobiko.ObjectNotFound:
            pass

        for connection in connections.with_attributes(failure=None):
            # connection not tried yet
            LOG.debug("Establishing SSH connection to "
                      f"'{connection.address}' (proxy_client={proxy_client})")
            try:
                ssh_client = self.ssh_client(connection.address,
                                             proxy_client=proxy_client,
                                             **connect_parameters)
                ssh_client.connect(retry_count=1, connection_attempts=1)
            except Exception as ex:
                LOG.debug("Failed establishing SSH connect to "
                          f"'{connection.address}': {ex}")
                # avoid re-checking again later the same address
                connection.failure = ex
                continue
            else:
                # cache valid connection SSH client for later use
                connection.ssh_client = ssh_client
                assert connection.is_valid
                return ssh_client

        failures = '\n'.join(str(connection.failure)
                             for connection in connections)
        raise UreachableSSHServer(addresses=addresses,
                                  failures=failures)

    def list_connections(
                self, addresses: typing.List[netaddr.IPAddress],
                proxy_client: typing.Optional[ssh.SSHClientFixture] = None
            ) -> typing.List[SSHConnection]:
        connections = []
        for address in addresses:
            connection = self.get_connection(address,
                                             proxy_client=proxy_client)
            connections.append(connection)
        return connections

    def get_connection(
                self, address: netaddr.IPAddress,
                proxy_client: typing.Optional[ssh.SSHClientFixture] = None
            ) -> SSHConnection:
        tobiko.check_valid_type(address, netaddr.IPAddress)
        if proxy_client is not None:
            tobiko.check_valid_type(proxy_client, ssh.SSHClientFixture)
        connection = SSHConnection(address, proxy_client=proxy_client)
        return self._connections.setdefault((address, proxy_client),
                                            connection)

    def ssh_client(self, address, username=None, port=None,
                   key_filename=None, **ssh_parameters):
        username = username or self.config.conf.username
        port = port or self.config.conf.port
        key_filename = key_filename or self.config.conf.key_file
        return ssh.ssh_client(host=str(address),
                              username=username,
                              key_filename=key_filename,
                              **ssh_parameters)


SSH_CONNECTIONS = SSHConnectionManager()


def ssh_connect(addresses: typing.List[netaddr.IPAddress],
                manager: typing.Optional[SSHConnectionManager] = None,
                **connect_parameters) -> ssh.SSHClientFixture:
    manager = manager or SSH_CONNECTIONS
    return manager.connect(addresses=addresses, **connect_parameters)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import stacks


class ServerGroupTestCase(testtools.TestCase):

    affinity_stack = tobiko.required_setup_fixture(
        stacks.AffinityServerGroupStackFixture)

    def test_affinity_server_group(self):
        group_id = self.affinity_stack.scheduler_group
        self.assertIsNotNone(group_id)

    anti_affinity_stack = tobiko.required_setup_fixture(
        stacks.AntiAffinityServerGroupStackFixture)

    def test_anti_affinity_server_group(self):
        group_id = self.anti_affinity_stack.scheduler_group
        self.assertIsNotNone(group_id)



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import os
import inspect
import sys

import fixtures
from oslo_log import log
import six
import testtools

import tobiko
from tobiko.common import _detail
from tobiko.common import _exception


LOG = log.getLogger(__name__)


def is_fixture(obj):
    '''Returns whenever obj is a fixture or not'''
    return (getattr(obj, '__tobiko_fixture__', False) or
            isinstance(obj, fixtures.Fixture) or
            (inspect.isclass(obj) and issubclass(obj, fixtures.Fixture)))


def get_fixture(obj, fixture_id=None, manager=None):
    '''Returns a fixture identified by given :param obj:

    It returns registered fixture for given :param obj:. If none has been
    registered it creates a new one.

    :param obj: can be:
      - an instance of fixtures.Fixture class: on such case it would return
        obj itself
      - the unique fully qualified name or an object that refers to a fixture
        class or an instance to a fixture.
      - the class of the fixture. It must be a subclass of fixtures.Fixture
        sub-class.

    :returns: an instance of fixture class identified by obj, or obj itself
    if it is instance of fixtures.Fixture class.

    '''
    if isinstance(obj, fixtures.Fixture):
        return obj
    else:
        manager = manager or FIXTURES
        return manager.get_fixture(obj, fixture_id=fixture_id)


def get_fixture_name(obj):
    '''Get unique fixture name'''
    name = getattr(obj, '__tobiko_fixture_name__', None)
    if name is None:
        if not is_fixture(obj):
            raise TypeError('Object {obj!r} is not a fixture.'.format(obj=obj))
        name = get_object_name(obj)
        obj.__tobiko_fixture__ = True
        obj.__tobiko_fixture_name__ = name
    return name


def get_fixture_class(obj):
    '''Get fixture class'''
    if isinstance(obj, six.string_types):
        obj = tobiko.load_object(obj)

    if not inspect.isclass(obj):
        obj = type(obj)

    assert issubclass(obj, fixtures.Fixture)
    return obj


def get_fixture_dir(obj):
    '''Get directory of fixture class source code file'''
    return os.path.dirname(inspect.getfile(get_fixture_class(obj)))


def remove_fixture(obj, fixture_id=None, manager=None):
    '''Unregister fixture identified by given :param obj: if any'''
    manager = manager or FIXTURES
    return manager.remove_fixture(obj, fixture_id=fixture_id)


def setup_fixture(obj, fixture_id=None, manager=None, alternative=None):
    '''Get registered fixture and setup it up'''
    if alternative is None:
        objs = [obj]
    else:
        objs = [obj, alternative]
    with _exception.handle_multiple_exceptions(
            handle_exception=handle_setup_error):
        errors = []
        for _obj in objs:
            fixture = get_fixture(_obj, fixture_id=fixture_id, manager=manager)
            try:
                fixture.setUp()
                break
            except testtools.MultipleExceptions:
                errors.append(sys.exc_info())
        else:
            raise testtools.MultipleExceptions(*errors)

    return fixture


def handle_setup_error(ex_type, ex_value, ex_tb):
    if issubclass(ex_type, fixtures.SetupError):
        details = ex_value.args[0]
        if details:
            details = {k: v.as_text() for k, v in details.items()}
            pretty_details = json.dumps(details, indent=4, sort_keys=True)
            LOG.debug(f"Fixture setup error details:\n{pretty_details}\n")
    else:
        LOG.exception("Unhandled setup exception",
                      exc_info=(ex_type, ex_value, ex_tb))


def reset_fixture(obj, fixture_id=None, manager=None):
    '''Get registered fixture and reset it'''
    fixture = get_fixture(obj, fixture_id=fixture_id, manager=manager)
    with _exception.handle_multiple_exceptions():
        fixture.reset()
    return fixture


def cleanup_fixture(obj, fixture_id=None, manager=None):
    '''Get registered fixture and clean it up'''
    fixture = get_fixture(obj, fixture_id=fixture_id, manager=manager)
    with _exception.handle_multiple_exceptions():
        fixture.cleanUp()
    return fixture


def get_name_and_object(obj):
    '''Get (name, obj) tuple identified by given :param obj:'''
    if isinstance(obj, six.string_types):
        return obj, tobiko.load_object(obj)
    else:
        return get_object_name(obj), obj


def visit_objects(objects):
    if not isinstance(objects, list):
        raise TypeError("parameter 'objects' is not a list")

    visited = set()
    while objects:
        obj = objects.pop()
        try:
            name, obj = get_name_and_object(obj)
        except Exception:
            LOG.exception('Unable to get (name, object) pair from {!r}'.format(
                obj))
        else:
            if name not in visited:
                visited.add(name)
                yield name, obj


def list_required_fixtures(objects):
    '''List fixture names required by given objects'''
    result = []
    objects = list(objects)
    for name, obj in visit_objects(objects):
        if is_fixture(obj):
            result.append(name)
            if not inspect.isclass(obj):
                obj = type(obj)

        elif is_test_method(obj):
            # Test methods also require test class fixtures
            if '.' in name:
                parent_name = name.rsplit('.', 1)[0]
                objects.append(parent_name)

        objects.extend(get_required_fixtures(obj))

    result.sort()
    return result


def is_test_method(obj):
    '''Returns whenever given object is a test method'''
    return ((inspect.isfunction(obj) or inspect.ismethod(obj)) and
            obj.__name__.startswith('test_'))


def get_required_fixtures(obj):
    '''Get fixture names required by given :param obj:'''
    required_names = getattr(obj, '__tobiko_required_fixtures__', None)
    if required_names is None:
        if is_test_method(obj):
            # Get fixtures from default values that are fixtures
            required = {default
                        for default in six.get_function_defaults(obj) or []
                        if is_fixture(default)}

        elif inspect.isclass(obj):
            # Get fixtures from members of type RequiredFixtureProperty
            required = {prop.fixture
                        for prop in get_required_fixture_properties(obj)}
        else:
            # Other types have no fixtures
            required = set()

        # Return every fixture name only once
        required_names = sorted([get_fixture_name(fixture)
                                 for fixture in required])
        try:
            # try to cache list for later use
            obj.__tobiko_required_fixtures__ = required_names
        except AttributeError:
            pass

    return required_names


def get_required_fixture_properties(cls):
    """Get list of members of type RequiredFixtureProperty of given class"""

    # inspect.getmembers() would iterate over such many testtools.TestCase
    # members too, so let exclude members from those very common base classes
    # that we know doesn't have members of type RequiredFixtureProperty
    base_classes = cls.__mro__
    for base_class in [testtools.TestCase, SharedFixture]:
        if issubclass(cls, base_class):
            base_classes = base_classes[:base_classes.index(base_class)]
            break

    # Get all members for selected class without calling properties or methods
    members = {}
    for base_class in reversed(base_classes):
        members.update(base_class.__dict__)

    # Return all members that are instances of RequiredFixtureProperty
    return [member
            for _, member in sorted(members.items())
            if isinstance(member, RequiredFixtureProperty)]


def init_fixture(obj, name):
    if inspect.isclass(obj) and issubclass(obj, fixtures.Fixture):
        try:
            obj = obj()
        except Exception as ex:
            raise TypeError(f"Error creating fixture '{name}' from class "
                            f"{obj!r}.") from ex
    if isinstance(obj, fixtures.Fixture):
        obj.__tobiko_fixture__ = True
        obj.__tobiko_fixture_name__ = name
        return obj

    raise TypeError(f"Invalid fixture object type: '{obj!r}'")


def fixture_property(*args, **kwargs):
    return FixtureProperty(*args, **kwargs)


def required_fixture(obj, **params):
    '''Creates a property that gets fixture identified by given :param obj:

    '''
    return RequiredFixtureProperty(obj, **params)


def required_setup_fixture(obj, **params):
    '''Creates a property that sets up fixture identified by given :param obj:

    '''
    return RequiredSetupFixtureProperty(obj, **params)


def get_object_name(obj):
    '''Gets a fully qualified name for given :param obj:'''
    if isinstance(obj, six.string_types):
        return obj

    name = getattr(obj, '__tobiko_fixture_name__', None)
    if name:
        return name

    if (not inspect.isfunction(obj) and
            not inspect.ismethod(obj) and
            not inspect.isclass(obj)):
        obj = type(obj)

    module = inspect.getmodule(obj).__name__

    if six.PY2:
        # Below code is only for old Python versions
        if inspect.isclass(obj):
            # This doesn't work for nested classes
            return module + '.' + obj.__name__

        method_class = getattr(obj, 'im_class', None)
        if method_class:
            # This doesn't work for nested classes
            return module + '.' + method_class.__name__ + '.' + obj.__name__

        if inspect.isfunction(obj):
            return module + '.' + obj.func_name

    else:
        # Only Python 3 defines __qualname__
        name = getattr(obj, '__qualname__', None)
        if name:
            return module + '.' + name

    msg = "Unable to get fixture name from object {!r}".format(obj)
    raise TypeError(msg)


class FixtureManager(object):

    def __init__(self):
        self.fixtures = {}

    def get_fixture(self, obj, fixture_id=None, init=None):
        name, obj = get_name_and_object(obj)
        if fixture_id:
            name += '-' + str(fixture_id)
        fixture = self.fixtures.get(name)
        if fixture is None:
            init = init or self.init_fixture
            self.fixtures[name] = fixture = init(obj=obj, name=name)
            assert isinstance(fixture, fixtures.Fixture)
        return fixture

    def init_fixture(self, obj, name):
        return init_fixture(obj=obj, name=name)

    def remove_fixture(self, obj, fixture_id=None):
        name = get_object_name(obj)
        if fixture_id:
            name += '-' + str(fixture_id)
        return self.fixtures.pop(name, None)


FIXTURES = FixtureManager()


class SharedFixture(fixtures.Fixture):
    """Base class for fixtures intended to be shared between multiple tests

    Make sure that fixture setUp method can be called more than once, but
    actually executing _setUp method only the first time. This allows the
    fixture to be passed to useFixture methods multiple times without caring
    about if has already been used before.

    Fixture set up can anyway be forced by calling 'setup_shared_fixture'
    method.

    Because cleanup policy in a shared fixture is different from a common
    fixture, cleanUp method simply doesn't nothing.

    Actual fixture cleanup is executed by calling
    cleanup_shared_fixture method.

    """

    _setup_executed = False
    _cleanup_executed = False

    __tobiko_fixture__ = True
    __tobiko_fixture_name__ = None

    def __init__(self):
        # make sure class states can be used before setUp
        self._clear_cleanups()

    @classmethod
    def get(cls, manager=None):
        return get_fixture(cls, manager=manager)

    def _remove_state(self):
        # make sure class states can be used after cleanUp
        super(SharedFixture, self)._clear_cleanups()

    def setUp(self):
        """Executes _setUp/setup_fixture method only the first time is called

        """
        if not self._setup_executed:
            LOG.debug('Set up fixture %r', self.fixture_name)
            super(SharedFixture, self).setUp()
            self._cleanup_executed = False
            self._setup_executed = True

    def cleanUp(self, raise_first=True):
        """Executes registered cleanups if any"""
        if not self._cleanup_executed:
            LOG.debug('Clean up fixture %r', self.fixture_name)
            self.addCleanup(self.cleanup_fixture)
        result = super(SharedFixture, self).cleanUp(raise_first=raise_first)
        self._setup_executed = False
        self._cleanup_executed = True
        return result

    def __enter__(self):
        return setup_fixture(self)

    def __exit__(self, exc_type, exc_val, exc_tb):  # noqa
        cleanup_fixture(self)
        return False  # propagate exceptions from the with body.

    def _setUp(self):
        self.setup_fixture()

    @property
    def fixture_name(self):
        return get_fixture_name(self)

    def setup_fixture(self):
        pass

    def cleanup_fixture(self):
        pass


class FixtureProperty(property):

    def __get__(self, instance, owner):
        instance = instance or tobiko.get_fixture(owner)
        return super(FixtureProperty, self).__get__(instance, owner)


class RequiredFixtureProperty(object):

    def __init__(self, fixture, **params):
        self.fixture = fixture
        self.fixture_params = params

    def __get__(self, instance, _):
        if instance is None:
            return self
        else:
            return self.get_fixture(instance)

    def get_fixture(self, _instance):
        return get_fixture(self.fixture, **self.fixture_params)

    @property
    def __tobiko_required_fixtures__(self):
        return [self.fixture]


class RequiredSetupFixtureProperty(RequiredFixtureProperty):

    def get_fixture(self, _instance):
        fixture = setup_fixture(self.fixture, **self.fixture_params)
        if (hasattr(_instance, 'addCleanup') and
                hasattr(_instance, 'getDetails')):
            _instance.addCleanup(_detail.gather_details,
                                 fixture.getDetails(),
                                 _instance.getDetails())
        return fixture



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


import os

from tobiko.common import _fixture


class TobikoConfigFicture(_fixture.SharedFixture):

    config = None

    def setup_fixture(self):
        from tobiko import config
        self.config = config.CONF.tobiko


def tobiko_config():
    return _fixture.setup_fixture(TobikoConfigFicture).config


def tobiko_config_dir():
    return tobiko_config().config_dir


def tobiko_config_path(path):
    path = os.path.expanduser(path)
    if not os.path.isabs(path):
        path = os.path.join(tobiko_config_dir(), path)
    return os.path.realpath(path)



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import itertools

import fixtures
from oslo_log import log
import six
import testtools
from testtools import content
import yaml


LOG = log.getLogger(__name__)


def gather_details(source_dict, target_dict):
    """Merge the details from ``source_dict`` into ``target_dict``.

    ``gather_details`` evaluates all details in ``source_dict``. Do not use it
    if the details are not ready to be evaluated.

    :param source_dict: A dictionary of details will be gathered.
    :param target_dict: A dictionary into which details will be gathered.
    """
    for name, content_object in source_dict.items():
        try:
            content_id = get_details_content_id(content_object)
            new_name = get_unique_detail_name(name=name,
                                              content_id=content_id,
                                              target_dict=target_dict)
            if new_name not in target_dict:
                target_dict[new_name] = copy_details_content(
                    content_object=content_object, content_id=content_id)
        except Exception:
            LOG.exception('Error gathering details')


def get_unique_detail_name(name, content_id, target_dict):
    disambiguator = itertools.count(1)
    new_name = name
    while new_name in target_dict:
        if content_id == get_details_content_id(target_dict[new_name]):
            break
        new_name = '{!s}-{!s}'.format(name, next(disambiguator))
    return new_name


testtools.testcase.gather_details = gather_details
fixtures.fixture.gather_details = gather_details


def copy_details_content(content_object, content_id):
    content_bytes = list(content_object.iter_bytes())
    return details_content(content_type=content_object.content_type,
                           get_bytes=lambda: content_bytes,
                           content_id=content_id)


def details_content(content_id, content_type=None, get_bytes=None,
                    get_text=None, get_json=None, get_yaml=None):
    content_type = content_type or content.UTF8_TEXT
    if get_bytes is None:
        if get_text:
            get_bytes = get_text_to_get_bytes(get_text=get_text)
        elif get_json:
            get_bytes = get_json_to_get_bytes(get_json=get_json)
        elif get_yaml:
            get_bytes = get_yaml_to_get_bytes(get_yaml=get_yaml)
        else:
            message = ("Any of get_bytes, get_text or get_json parameters has "
                       "been specified")
            raise ValueError(message)
    content_object = content.Content(content_type=content_type,
                                     get_bytes=get_bytes)
    content_object.content_id = content_id
    return content_object


def get_text_to_get_bytes(get_text):
    assert callable(get_text)

    def get_bytes():
        text = get_text()
        if text:
            if isinstance(text, six.string_types):
                yield text.encode(errors='ignore')
            else:
                for t in text:
                    yield t.encode(errors='ignore')

    return get_bytes


def get_json_to_get_bytes(get_json):
    assert callable(get_json)

    def get_text():
        obj = get_json()
        yield json.dumps(obj, indent=4, sort_keys=True).encode(errors='ignore')

    return get_text


def get_yaml_to_get_bytes(get_yaml):
    assert callable(get_yaml)

    def get_text():
        obj = get_yaml()
        yield yaml.dump(obj).encode(errors='ignore')

    return get_text


def get_details_content_id(content_object):
    try:
        return content_object.content_id
    except AttributeError:
        return id(content_object)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from heatclient.v1 import client as heatclient

import tobiko
from tobiko.openstack import _client


class HeatClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session):
        return heatclient.Client(session=session,
                                 endpoint_type='public',
                                 service_type='orchestration')


class HeatClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return HeatClientFixture(session=session)


CLIENTS = HeatClientManager()


def heat_client(obj=None):
    if obj is None:
        return default_heat_client()

    if isinstance(obj, heatclient.Client):
        return obj

    fixture = tobiko.get_fixture(obj)
    if isinstance(fixture, HeatClientFixture):
        return tobiko.setup_fixture(fixture).client

    message = "Object {!r} is not a NeutronClientFixture".format(obj)
    raise TypeError(message)


def default_heat_client():
    return get_heat_client()


def get_heat_client(session=None, shared=True, init_client=None,
                    manager=None):
    manager = manager or CLIENTS
    fixture = manager.get_client(session=session, shared=shared,
                                 init_client=init_client)
    return tobiko.setup_fixture(fixture).client



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import netaddr
import testtools

import tobiko
from tobiko.openstack import stacks
from tobiko.shell import ip
from tobiko.shell import ssh
from tobiko.tests.functional.shell import _fixtures


class IpTest(testtools.TestCase):

    centos_stack = tobiko.required_setup_fixture(
        stacks.CentosServerStackFixture)

    cirros_stack = tobiko.required_setup_fixture(
        stacks.CirrosServerStackFixture)

    ubuntu_stack = tobiko.required_setup_fixture(
        stacks.UbuntuServerStackFixture)

    namespace = tobiko.required_setup_fixture(
        _fixtures.NetworkNamespaceFixture)

    def test_list_ip_addresses(self, ip_version=None, scope=None,
                               **execute_params):
        if not os.path.isfile('/bin/ip'):
            self.skip("'bin/ip' command not found")
        ips = ip.list_ip_addresses(ip_version=ip_version, scope=scope,
                                   **execute_params)
        self.assertIsInstance(ips, tobiko.Selection)
        for ip_address in ips:
            self.assertIsInstance(ip_address, netaddr.IPAddress)
        if ip_version:
            self.assertEqual(ips.with_attributes(version=ip_version), ips)
        if scope:
            if scope == 'link':
                self.assertEqual(ips.with_attributes(version=4), [])
                self.assertEqual(ips.with_attributes(version=6), ips)
            elif scope == 'host':
                for a in ips:
                    self.assertTrue(a.is_loopback())
            elif scope == 'global':
                self.assertNotIn(netaddr.IPAddress('127.0.0.1'), ips)
                self.assertNotIn(netaddr.IPAddress('::1'), ips)
        return ips

    def test_list_ip_addresses_with_host_scope(self, **execute_params):
        self.test_list_ip_addresses(scope='host', **execute_params)

    def test_list_ip_addresses_with_link_scope(self, **execute_params):
        self.test_list_ip_addresses(scope='link', **execute_params)

    def test_list_ip_addresses_with_global_scope(self, **execute_params):
        self.test_list_ip_addresses(scope='global', **execute_params)

    def test_list_ip_addresses_with_ipv4(self):
        self.test_list_ip_addresses(ip_version=4)

    def test_list_ip_addresses_with_ipv6(self):
        self.test_list_ip_addresses(ip_version=6)

    def test_list_ip_addresses_with_centos_server(self):
        self.test_list_ip_addresses(ssh_client=self.centos_stack.ssh_client)

    def test_list_ip_addresses_with_cirros_server(self):
        self.test_list_ip_addresses(ssh_client=self.cirros_stack.ssh_client)

    def test_list_ip_addresses_with_ubuntu_server(self):
        self.test_list_ip_addresses(ssh_client=self.ubuntu_stack.ssh_client)

    def test_list_ip_addresses_with_proxy_ssh_client(self):
        ssh_client = ssh.ssh_proxy_client()
        if ssh_client is None:
            self.skip('SSH proxy server not configured')
        self.test_list_ip_addresses(ssh_client=ssh_client)

    def test_list_ip_addresses_with_proxy_ssh_client_and_host_scope(
                self, **execute_params):
        self.test_list_ip_addresses(scope='host', **execute_params)

    def test_list_ip_addresses_with_proxy_ssh_client_and_link_scope(
                self, **execute_params):
        self.test_list_ip_addresses(scope='link', **execute_params)

    def test_list_ip_addresses_with_proxy_ssh_client_and_global_scope(
                self, **execute_params):
        self.test_list_ip_addresses(scope='global', **execute_params)

    def test_list_ip_addresses_with_namespace(self, **params):
        namespace_ips = ip.list_ip_addresses(
            ssh_client=self.namespace.ssh_client,
            network_namespace=self.namespace.network_namespace, **params)
        self.assertNotEqual([], namespace_ips)

        host_ips = ip.list_ip_addresses(ssh_client=self.namespace.ssh_client,
                                        **params)
        self.assertNotEqual(host_ips, namespace_ips)

    def test_list_ip_addresses_with_namespace_and_scope(self):
        self.test_list_ip_addresses_with_namespace(scope='global')

    def test_list_ip_addresses_with_failing_command(self):
        self.assertRaises(ip.IpError, ip.list_ip_addresses,
                          ip_command=['false'],
                          ssh_client=self.namespace.ssh_client)

    def test_list_ip_addresses_with_ignore_errors(self, **execute_params):
        self.test_list_ip_addresses(ignore_errors=True, ip_command='false',
                                    **execute_params)

    def test_list_namespaces(self, **execute_params):
        if not os.path.isfile('/bin/ip'):
            self.skip("'bin/ip' command not found")
        namespaces = ip.list_network_namespaces(**execute_params)
        self.assertIsInstance(namespaces, list)
        for namespace in namespaces:
            self.assertIsInstance(namespace, str)
            self.test_list_ip_addresses(network_namespace=namespace)

    def test_list_namespaces_with_centos_server(self):
        self.test_list_namespaces(ssh_client=self.centos_stack.ssh_client)

    def test_list_namespaces_with_ubuntu_server(self):
        self.test_list_namespaces(ssh_client=self.ubuntu_stack.ssh_client)

    def test_list_namespaces_with_proxy_ssh_client(self):
        ssh_client = ssh.ssh_proxy_client()
        if ssh_client is None:
            self.skip('SSH proxy server not configured')
        self.test_list_namespaces(ssh_client=ssh_client)

    def test_list_namespaces_with_failing_command(self):
        self.assertRaises(ip.IpError, ip.list_network_namespaces,
                          ip_command=['false'],
                          ssh_client=self.namespace.ssh_client)

    def test_list_namespaces_with_ignore_errors(self, **execute_params):
        self.test_list_namespaces(ignore_errors=True, ip_command='false',
                                  **execute_params)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack.stacks import _centos


CONF = config.CONF

RHEL_IMAGE_MAJOR_VERSION = '8.2'
RHEL_IMAGE_MINOR_VERSION = '501'

RHEL_IMAGE_URL = ('http://download.devel.redhat.com/brewroot/packages/'
                  f'rhel-guest-image/{RHEL_IMAGE_MAJOR_VERSION}/'
                  f'{RHEL_IMAGE_MINOR_VERSION}/images/'
                  f'rhel-guest-image-{RHEL_IMAGE_MAJOR_VERSION}-'
                  f'{RHEL_IMAGE_MINOR_VERSION}.x86_64.qcow2')


class RhelImageFixture(glance.URLGlanceImageFixture):

    image_url = CONF.tobiko.rhel.image_url or RHEL_IMAGE_URL
    image_name = CONF.tobiko.rhel.image_name
    image_file = CONF.tobiko.rhel.image_file
    disk_format = CONF.tobiko.rhel.disk_format or "qcow2"
    container_format = CONF.tobiko.rhel.container_format or "bare"
    username = CONF.tobiko.rhel.username or 'cloud-user'
    password = CONF.tobiko.rhel.password
    connection_timeout = CONF.tobiko.rhel.connection_timeout


class RedHatFlavorStackFixture(_centos.CentosFlavorStackFixture):
    pass


class RedHatServerStackFixture(_centos.CentosServerStackFixture):

    #: Glance image used to create a Nova server instance
    # (alternative is given for cases the RHEL image is failed to be
    # set up)
    image_fixture = tobiko.required_setup_fixture(
        RhelImageFixture, alternative=_centos.CentosImageFixture)

    #: Flavor used to create a Nova server instance
    flavor_stack = tobiko.required_setup_fixture(RedHatFlavorStackFixture)



from __future__ import absolute_import

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import topology
from tobiko.shell import ip


@keystone.skip_unless_has_keystone_credentials()
class NetworkNamespaceFixture(tobiko.SharedFixture):

    network_namespace = None
    ssh_client = None

    def setup_fixture(self):
        for node in topology.list_openstack_nodes():
            network_namespaces = ip.list_network_namespaces(
                ignore_errors=True,
                ssh_client=node.ssh_client)
            if network_namespaces:
                self.network_namespace = network_namespaces.first
                self.ssh_client = node.ssh_client



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr
import testtools

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.openstack import tests


CONF = config.CONF


@keystone.skip_unless_has_keystone_credentials()
class NeutronApiTest(testtools.TestCase):
    """Tests network creation"""

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_setup_fixture(stacks.NetworkStackFixture)

    def test_find_network_with_id(self):
        network = neutron.find_network(id=self.stack.network_id)
        self.assertEqual(self.stack.network_id, network['id'])

    def test_find_floating_network(self):
        floating_network = CONF.tobiko.neutron.floating_network
        if not floating_network:
            tobiko.skip_test('floating_network not configured')
        network = neutron.find_network(name=floating_network)
        self.assertIn(floating_network, [network['name'], network['id']])
        self.assertEqual(self.stack.gateway_network_id, network['id'])

    def test_list_networks(self):
        networks = neutron.list_networks()
        network_ids = {n['id'] for n in networks}
        self.assertIn(self.stack.network_id, network_ids)

    def test_list_subnets(self):
        subnets = neutron.list_subnets()
        subnets_ids = {s['id'] for s in subnets}
        if self.stack.has_ipv4:
            self.assertIn(self.stack.ipv4_subnet_id, subnets_ids)
        if self.stack.has_ipv6:
            self.assertIn(self.stack.ipv6_subnet_id, subnets_ids)

    def test_list_subnet_cidrs(self):
        subnets_cidrs = neutron.list_subnet_cidrs()
        if self.stack.has_ipv4:
            cidr = netaddr.IPNetwork(self.stack.ipv4_subnet_details['cidr'])
            self.assertIn(cidr, subnets_cidrs)
        if self.stack.has_ipv6:
            cidr = netaddr.IPNetwork(self.stack.ipv6_subnet_details['cidr'])
            self.assertIn(cidr, subnets_cidrs)

    def test_get_network(self):
        network = neutron.get_network(self.stack.network_id)
        self.assertEqual(self.stack.network_id, network['id'])
        self.assertEqual(self.stack.port_security_enabled,
                         network['port_security_enabled'])
        if self.stack.has_ipv4:
            self.assertIn(self.stack.ipv4_subnet_id, network['subnets'])
        else:
            self.assertNotIn(self.stack.ipv4_subnet_id, network['subnets'])
        if self.stack.has_ipv6:
            self.assertIn(self.stack.ipv6_subnet_id, network['subnets'])
        else:
            self.assertNotIn(self.stack.ipv6_subnet_id, network['subnets'])

    def test_create_network(self):
        network = neutron.create_network(name=self.id())
        self.addCleanup(neutron.delete_network, network['id'])
        self.assertIsInstance(network['id'], str)
        self.assertNotEqual('', network['id'])
        self.assertEqual(self.id(), network['name'])
        observed = neutron.get_network(network['id'])
        self.assertEqual(network['id'], observed['id'])

    def test_delete_network(self):
        network = neutron.create_network(name=self.id())
        neutron.delete_network(network['id'])
        self.assertRaises(neutron.NoSuchNetwork, neutron.get_network,
                          network['id'])

    def test_get_router(self):
        if not self.stack.has_gateway:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no gateway "
                             "router")
        router = neutron.get_router(self.stack.gateway_id)
        self.assertEqual(self.stack.gateway_id, router['id'])

    def test_get_ipv4_subnet(self):
        if not self.stack.has_ipv4:
            tobiko.skip_test(
                "Stack {self.stack.stack_name} has no IPv4 subnet")
        subnet = neutron.get_subnet(self.stack.ipv4_subnet_id)
        self.assertEqual(self.stack.ipv4_subnet_id, subnet['id'])
        self.assertEqual(self.stack.ipv4_subnet_details, subnet)

    def test_get_ipv6_subnet(self):
        if not self.stack.has_ipv6:
            tobiko.skip_test(
                "Stack {self.stack.stack_name} has no IPv6 subnet")
        subnet = neutron.get_subnet(self.stack.ipv6_subnet_id)
        self.assertEqual(self.stack.ipv6_subnet_id, subnet['id'])
        self.assertEqual(self.stack.ipv6_subnet_details, subnet)

    def test_find_agents_with_binary(self):
        agent = neutron.list_agents().first
        agents = neutron.list_agents(binary=agent['binary'])
        self.assertIn(agent['id'], {a['id'] for a in agents})


@keystone.skip_unless_has_keystone_credentials()
class PortTest(testtools.TestCase):

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_setup_fixture(stacks.CentosServerStackFixture)

    def test_list_port_addresses(self, ip_version=None):
        port = neutron.find_port(device_id=self.stack.server_id)
        port_addresses = neutron.list_port_ip_addresses(
            port=port,
            ip_version=ip_version)
        server_addresses = nova.list_server_ip_addresses(
            server=self.stack.server_details,
            ip_version=ip_version,
            address_type='fixed')
        self.assertEqual(set(server_addresses), set(port_addresses))
        if ip_version:
            self.assertEqual(
                port_addresses.with_attributes(version=ip_version),
                port_addresses)

    def test_list_port_addresses_with_ipv4(self):
        self.test_list_port_addresses(ip_version=4)

    def test_list_port_addresses_with_ipv6(self):
        self.test_list_port_addresses(ip_version=6)

    def test_find_port_address_with_ip_version(self):
        port = neutron.find_port(device_id=self.stack.server_id)
        server_addresses = nova.list_server_ip_addresses(
            server=self.stack.server_details,
            address_type='fixed')
        for server_address in server_addresses:
            port_address = neutron.find_port_ip_address(
                port=port,
                ip_version=server_address.version,
                unique=True)
            self.assertEqual(server_address, port_address)

    def test_find_port_address_with_subnet_id(self):
        port = neutron.find_port(device_id=self.stack.server_id)
        for subnet in neutron.list_subnets(network_id=port['network_id']):
            port_address = neutron.find_port_ip_address(
                port=port, subnet_id=subnet['id'], unique=True)
            cidr = netaddr.IPNetwork(subnet['cidr'])
            self.assertIn(port_address, cidr)


@keystone.skip_unless_has_keystone_credentials()
class AgentTest(testtools.TestCase):

    def test_skip_if_missing_agents(self, count=1, should_skip=False,
                                    **params):
        if should_skip:
            expected_exeption = self.skipException
        else:
            expected_exeption = self.failureException

        @neutron.skip_if_missing_networking_agents(count=count, **params)
        def method():
            raise self.fail('Not skipped')

        exception = self.assertRaises(expected_exeption, method)
        if should_skip:
            agents = neutron.list_agents(**params)
            message = "missing {!r} agent(s)".format(count - len(agents))
            if params:
                message += " with {!s}".format(
                    ','.join('{!s}={!r}'.format(k, v)
                             for k, v in params.items()))
            self.assertEqual(message, str(exception))
        else:
            self.assertEqual('Not skipped', str(exception))

    def test_skip_if_missing_agents_with_no_agents(self):
        self.test_skip_if_missing_agents(binary='never-never-land',
                                         should_skip=True)

    def test_skip_if_missing_agents_with_big_count(self):
        self.test_skip_if_missing_agents(count=1000000,
                                         should_skip=True)

    def test_neutron_agents_are_alive(self):
        agents = tests.test_neutron_agents_are_alive()
        # check has agents and they are all alive
        self.assertNotEqual([], agents)
        self.assertNotEqual([], agents.with_items(alive=True))



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import typing  # noqa

import six
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import heat
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack.stacks import _hot
from tobiko.openstack.stacks import _neutron
from tobiko.shell import ssh
from tobiko.shell import sh


CONF = config.CONF
LOG = log.getLogger(__name__)


class KeyPairStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('nova/key_pair.yaml')
    key_file = tobiko.tobiko_config_path(CONF.tobiko.nova.key_file)
    public_key = None
    private_key = None

    def setup_fixture(self):
        self.create_key_file()
        self.read_keys()
        super(KeyPairStackFixture, self).setup_fixture()

    def read_keys(self):
        with open(self.key_file, 'r') as fd:
            self.private_key = as_str(fd.read())
        with open(self.key_file + '.pub', 'r') as fd:
            self.public_key = as_str(fd.read())

    def create_key_file(self):
        key_file = self.key_file
        if not os.path.isfile(key_file):
            key_dir = os.path.dirname(key_file)
            tobiko.makedirs(key_dir)
            try:
                sh.local_execute(['ssh-keygen', '-f', key_file, '-P', ''])
            except sh.ShellCommandFailed:
                if not os.path.isfile(key_file):
                    raise
            else:
                assert os.path.isfile(key_file)


class FlavorStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('nova/flavor.yaml')

    disk: typing.Optional[int] = None
    ephemeral = None
    extra_specs = None
    is_public = None
    name = None
    rxtx_factor = None
    swap = None
    vcpus = None


@neutron.skip_if_missing_networking_extensions('port-security')
class ServerStackFixture(heat.HeatStackFixture):

    #: Heat template file
    template = _hot.heat_template_file('nova/server.yaml')

    #: stack with the key pair for the server instance
    key_pair_stack = tobiko.required_setup_fixture(KeyPairStackFixture)

    #: stack with the internal where the server port is created
    network_stack = tobiko.required_setup_fixture(_neutron.NetworkStackFixture)

    #: Glance image used to create a Nova server instance
    image_fixture = None

    def delete_stack(self, stack_id=None):
        if self._outputs:
            tobiko.cleanup_fixture(self.ssh_client)
        super(ServerStackFixture, self).delete_stack(stack_id=stack_id)

    @property
    def image(self):
        return self.image_fixture.image_id

    @property
    def username(self):
        """username used to login to a Nova server instance"""
        return self.image_fixture.username

    @property
    def password(self):
        """password used to login to a Nova server instance"""
        return self.image_fixture.password

    @property
    def connection_timeout(self):
        return self.image_fixture.connection_timeout

    # Stack used to create flavor for Nova server instance
    flavor_stack = None

    @property
    def flavor(self):
        """Flavor for Nova server instance"""
        return self.flavor_stack.flavor_id

    #: Whenever port security on internal network is enable
    port_security_enabled = False

    #: Security groups to be associated to network ports
    security_groups = []  # type: typing.List[str]

    @property
    def key_name(self):
        return self.key_pair_stack.key_name

    @property
    def network(self):
        return self.network_stack.network_id

    #: Floating IP network where the Neutron floating IP are created
    @property
    def floating_network(self):
        return self.network_stack.floating_network

    @property
    def has_floating_ip(self):
        return bool(self.floating_network)

    @property
    def ssh_client(self) -> ssh.SSHClientFixture:
        return ssh.ssh_client(host=self.ip_address,
                              username=self.username,
                              password=self.password,
                              connection_timeout=self.connection_timeout)

    @property
    def ssh_command(self):
        return ssh.ssh_command(host=self.ip_address,
                               username=self.username)

    @property
    def ip_address(self):
        if self.has_floating_ip:
            return self.floating_ip_address
        else:
            return self.outputs.fixed_ips[0]['ip_address']

    #: Schedule on different host that this Nova server instance ID
    different_host = None

    #: Schedule on same host as this Nova server instance ID
    same_host = None

    #: Scheduler group in which this Nova server is attached
    @property
    def scheduler_group(self):
        return None

    @property
    def scheduler_hints(self):
        scheduler_hints = {}
        if self.different_host:
            scheduler_hints.update(different_host=list(self.different_host))
        if self.same_host:
            scheduler_hints.update(same_host=list(self.same_host))
        if self.scheduler_group:
            scheduler_hints.update(group=self.scheduler_group)
        return scheduler_hints

    #: allow to retry creating server in case scheduler hits are not respected
    retry_create = 3
    expected_creted_status = {heat.CREATE_COMPLETE}

    def validate_created_stack(self):
        stack = super(ServerStackFixture, self).validate_created_stack()
        self.validate_scheduler_hints()
        return stack

    @property
    def hypervisor_host(self):
        return getattr(self.server_details, 'OS-EXT-SRV-ATTR:host')

    def validate_scheduler_hints(self):
        if self.scheduler_hints:
            hypervisor = self.hypervisor_host
            self.validate_same_host_scheduler_hints(hypervisor=hypervisor)
            self.validate_different_host_scheduler_hints(hypervisor=hypervisor)

    def validate_same_host_scheduler_hints(self, hypervisor):
        if self.same_host:
            different_host_hypervisors = nova.get_different_host_hypervisors(
                self.same_host, hypervisor)
            if different_host_hypervisors:
                tobiko.skip_test(f"Server {self.server_id} of stack "
                                 f"{self.stack_name} created on different "
                                 "hypervisor host from servers:\n"
                                 f"{different_host_hypervisors}")

    def validate_different_host_scheduler_hints(self, hypervisor):
        if self.different_host:
            same_host_hypervisors = nova.get_same_host_hypervisors(
                self.different_host, hypervisor)
            if same_host_hypervisors:
                tobiko.skip_test(f"Server {self.server_id} of stack "
                                 f"{self.stack_name} created on the same "
                                 "hypervisor host as servers:\n"
                                 f"{same_host_hypervisors}")

    @property
    def server_details(self):
        return nova.get_server(self.server_id)

    @property
    def port_details(self):
        return neutron.get_port(self.port_id)

    def getDetails(self):
        # pylint: disable=W0212
        details = super(ServerStackFixture, self).getDetails()
        stack = self.get_stack()
        if stack:
            details[self.fixture_name + '.stack'] = (
                self.details_content(get_json=lambda: stack._info))
            if stack.stack_status == 'CREATE_COMPLETE':
                details[self.fixture_name + '.server_details'] = (
                    self.details_content(
                        get_json=lambda: self.server_details._info))
                details[self.fixture_name + '.console_output'] = (
                    self.details_content(
                        get_text=lambda: self.console_output))
        return details

    def details_content(self, **kwargs):
        return tobiko.details_content(content_id=self.fixture_name, **kwargs)

    max_console_output_length = 64 * 1024

    @property
    def console_output(self):
        return nova.get_console_output(server=self.server_id,
                                       length=self.max_console_output_length)

    @property
    def user_data(self):
        return nova.user_data(self.cloud_config)

    #: SWAP file name
    swap_filename: str = '/swap.img'
    #: SWAP file size in bytes
    swap_size: typing.Optional[int] = None
    #: nax SWAP file size in bytes
    swap_maxsize: typing.Optional[int] = None

    @property
    def cloud_config(self):
        cloud_config = nova.cloud_config()
        # default is to not create any swap files,
        # because 'swap_file_max_size' is set to None
        if self.swap_maxsize is not None:
            cloud_config = nova.cloud_config(
                cloud_config,
                swap={'filename': self.swap_filename,
                      'size': self.swap_size or 'auto',
                      'maxsize': self.swap_maxsize})
        return cloud_config

    def ensure_server_status(self, status):
        tobiko.setup_fixture(self)
        try:
            server = nova.wait_for_server_status(self.server_id, status)
        except nova.WaitForServerStatusError:
            server = nova.get_server(self.server_id)
            LOG.debug(f"Server {server.id} status is {server.status} instead "
                      f"of {status}", exc_info=1)
        if server.status == status:
            return server
        elif status == "ACTIVE":
            tobiko.reset_fixture(self)
            return nova.wait_for_server_status(self.server_id, 'ACTIVE')
        else:
            tobiko.skip_test(f"{type(self).__name__}.ensure_server_status "
                             "method not implemented")


class ExternalServerStackFixture(ServerStackFixture):

    #: stack with the network where the server port is created
    network_stack = tobiko.required_setup_fixture(
        _neutron.ExternalNetworkStackFixture)

    # external servers doesn't need floating IPs
    has_floating_ip = False

    @property
    def floating_network(self):
        return self.network_stack.network_id


class PeerServerStackFixture(ServerStackFixture):
    """Server witch networking access requires passing by another Nova server
    """

    has_floating_ip = False

    #: Peer server used to reach this one
    peer_stack = None

    @property
    def ssh_client(self):
        return ssh.ssh_client(host=self.ip_address,
                              username=self.username,
                              password=self.password,
                              connection_timeout=self.connection_timeout,
                              proxy_jump=self.peer_stack.ssh_client)

    @property
    def ssh_command(self):
        proxy_command = self.peer_stack.ssh_command + [
            'nc', self.ip_address, '22']
        return ssh.ssh_command(host=self.ip_address,
                               username=self.username,
                               proxy_command=proxy_command)

    @property
    def network(self):
        return self.peer_stack.network


@nova.skip_if_missing_hypervisors(count=2, state='up', status='enabled')
class DifferentHostServerStackFixture(PeerServerStackFixture):

    @property
    def different_host(self):
        return [self.peer_stack.server_id]


class SameHostServerStackFixture(PeerServerStackFixture):

    @property
    def same_host(self):
        return [self.peer_stack.server_id]


def as_str(text):
    if isinstance(text, six.string_types):
        return text
    else:
        return text.decode()


class ServerGroupStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('nova/server_group.yaml')


class AffinityServerGroupStackFixture(tobiko.SharedFixture):
    server_group_stack = tobiko.required_setup_fixture(
        ServerGroupStackFixture)

    @property
    def scheduler_group(self):
        return self.server_group_stack.affinity_server_group_id


class AntiAffinityServerGroupStackFixture(tobiko.SharedFixture):
    server_group_stack = tobiko.required_setup_fixture(
        ServerGroupStackFixture)

    @property
    def scheduler_group(self):
        return self.server_group_stack.anti_affinity_server_group_id



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import typing

import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import heat
from tobiko.openstack import neutron
from tobiko.openstack.stacks import _hot
from tobiko.shell import ip
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


CONF = config.CONF
LOG = log.getLogger(__name__)

NeutronNetworkType = typing.Dict[str, typing.Any]


class ExternalNetworkStackFixture(heat.HeatStackFixture):

    template = _hot.heat_template_file('neutron/external_network.yaml')

    @property
    def external_name(self):
        return tobiko.tobiko_config().neutron.external_network

    subnet_enable_dhcp: typing.Optional[bool] = True

    _external_network: typing.Optional[NeutronNetworkType] = None

    @property
    def external_network(self) -> typing.Optional[NeutronNetworkType]:
        external_network = self._external_network
        if external_network is None:
            subnet_parameters = {}
            if self.subnet_enable_dhcp is not None:
                subnet_parameters['enable_dhcp'] = self.subnet_enable_dhcp
            for network in list_external_networks(name=self.external_name):
                if not network['subnets']:
                    LOG.debug(f"Network '{network['id']}' has any subnet")
                    continue
                if subnet_parameters:
                    subnets = neutron.list_subnets(network_id=network['id'],
                                                   **subnet_parameters)
                    if not subnets:
                        LOG.debug(f"Network '{network['id']}' has any valid "
                                  f"subnet: {subnet_parameters}")
                        continue

                network_dump = json.dumps(network, indent=4, sort_keys=True)
                LOG.debug(f"Found external network for {self.fixture_name}:\n"
                          f"{network_dump}")
                self._external_network = external_network = network
                break
            else:
                LOG.warning("No external network found for "
                            f"'{self.fixture_name}':\n"
                            f" - name or ID: {self.external_name}\n"
                            f" - subnet attributes: {subnet_parameters}\n")
        return external_network

    @property
    def external_id(self):
        network = self.external_network
        return network and network['id'] or None

    @property
    def has_external_id(self):
        return bool(self.external_id)

    @property
    def network_details(self):
        return neutron.get_network(self.network_id)

    has_gateway = False


class FloatingNetworkStackFixture(ExternalNetworkStackFixture):

    @property
    def external_name(self):
        return tobiko.tobiko_config().neutron.floating_network

    subnet_enable_dhcp = None


@neutron.skip_if_missing_networking_extensions('port-security')
class NetworkStackFixture(heat.HeatStackFixture):
    """Heat stack for creating internal network with a router to external"""

    #: Heat template file
    template = _hot.heat_template_file('neutron/network.yaml')

    #: Disable port security by default for new network ports
    port_security_enabled = False

    @property
    def has_ipv4(self):
        """Whenever to setup IPv4 subnet"""
        return bool(CONF.tobiko.neutron.ipv4_cidr)

    @property
    def ipv4_cidr(self):
        if self.has_ipv4:
            return neutron.new_ipv4_cidr(seed=self.fixture_name)
        else:
            return None

    @property
    def has_ipv6(self):
        """Whenever to setup IPv6 subnet"""
        return bool(CONF.tobiko.neutron.ipv6_cidr)

    @property
    def ipv6_cidr(self):
        if self.has_ipv6:
            return neutron.new_ipv6_cidr(seed=self.fixture_name)
        else:
            return None

    @property
    def network_value_specs(self):
        """Extra network creation parameters"""
        return {}

    floating_network_stack = tobiko.required_setup_fixture(
        FloatingNetworkStackFixture)

    @property
    def floating_network(self):
        """Network ID where the Neutron floating IPs are created"""
        return self.floating_network_stack.network_id

    @property
    def gateway_network(self):
        """Network ID where gateway routes packages to"""
        return self.floating_network

    ha = False

    @property
    def gateway_value_specs(self):
        value_specs = {}
        if self.has_l3_ha:
            value_specs.update(ha=(self.ha or False))
        return value_specs

    @property
    def has_gateway(self):
        """Whenever to setup gateway router"""
        return bool(self.gateway_network)

    @property
    def has_net_mtu(self):
        """Whenever can obtain network MTU value"""
        return neutron.has_networking_extensions('net-mtu')

    @property
    def has_l3_ha(self):
        """Whenever can obtain gateway router HA value"""
        return neutron.has_networking_extensions('l3-ha')

    @property
    def network_details(self):
        return neutron.get_network(self.network_id)

    @property
    def ipv4_subnet_details(self):
        return neutron.get_subnet(self.ipv4_subnet_id)

    @property
    def ipv4_subnet_cidr(self):
        return netaddr.IPNetwork(self.ipv4_subnet_details['cidr'])

    @property
    def ipv4_subnet_gateway_ip(self):
        return netaddr.IPAddress(self.ipv4_subnet_details['gateway_ip'])

    @property
    def ipv4_dns_nameservers(self):
        nameservers = CONF.tobiko.neutron.ipv4_dns_nameservers
        if nameservers is None:
            nameservers = default_nameservers(ip_version=4)
        return ','.join(str(nameserver) for nameserver in nameservers)

    @property
    def ipv6_subnet_details(self):
        return neutron.get_subnet(self.ipv6_subnet_id)

    @property
    def ipv6_subnet_cidr(self):
        return netaddr.IPNetwork(self.ipv6_subnet_details['cidr'])

    @property
    def ipv6_subnet_gateway_ip(self):
        return netaddr.IPAddress(self.ipv6_subnet_details['gateway_ip'])

    @property
    def ipv6_dns_nameservers(self):
        nameservers = CONF.tobiko.neutron.ipv6_dns_nameservers
        if nameservers is None:
            nameservers = default_nameservers(ip_version=6)
        return ','.join(str(nameserver) for nameserver in nameservers)

    @property
    def gateway_details(self):
        return neutron.get_router(self.gateway_id)

    @property
    def external_gateway_ips(self):
        fixed_ips = self.gateway_details['external_gateway_info'][
            'external_fixed_ips']
        return tobiko.select(netaddr.IPAddress(fixed_ip['ip_address'])
                             for fixed_ip in fixed_ips)

    @property
    def ipv4_gateway_ports(self):
        return neutron.list_ports(fixed_ips='subnet_id=' + self.ipv4_subnet_id,
                                  device_id=self.gateway_id,
                                  network_id=self.network_id)

    @property
    def ipv6_gateway_ports(self):
        return neutron.list_ports(fixed_ips='subnet_id=' + self.ipv6_subnet_id,
                                  device_id=self.gateway_id,
                                  network_id=self.network_id)

    @property
    def external_geteway_ports(self):
        return neutron.list_ports(device_id=self.gateway_id,
                                  network_id=self.gateway_network_id)

    @property
    def ipv4_gateway_addresses(self):
        ips = tobiko.Selection()
        for port in self.ipv4_gateway_ports:
            ips.extend(neutron.list_port_ip_addresses(port))
        return ips

    @property
    def ipv6_gateway_addresses(self):
        ips = tobiko.Selection()
        for port in self.ipv6_gateway_ports:
            ips.extend(neutron.list_port_ip_addresses(port))
        return ips

    @property
    def external_gateway_addresses(self):
        ips = tobiko.Selection()
        for port in self.external_geteway_ports:
            ips.extend(neutron.list_port_ip_addresses(port))
        return ips

    @property
    def gateway_network_details(self):
        return neutron.get_network(self.gateway_network_id)


@neutron.skip_if_missing_networking_extensions('net-mtu-writable')
class NetworkWithNetMtuWriteStackFixture(NetworkStackFixture):

    @property
    def custom_mtu_size(self):
        return CONF.tobiko.neutron.custom_mtu_size

    @property
    def network_value_specs(self):
        value_specs = super(NetworkWithNetMtuWriteStackFixture,
                            self).network_value_specs
        return dict(value_specs, mtu=self.custom_mtu_size)


@neutron.skip_if_missing_networking_extensions('security-group')
class SecurityGroupsFixture(heat.HeatStackFixture):
    """Heat stack with some security groups

    """
    #: Heat template file
    template = _hot.heat_template_file('neutron/security_groups.yaml')


def list_external_networks(name: typing.Optional[str] = None) -> \
        tobiko.Selection[NeutronNetworkType]:
    networks = tobiko.Selection[NeutronNetworkType]()
    if name is not None:
        try:
            network = neutron.get_network(name)
        except neutron.NoSuchNetwork:
            LOG.error(f"No such network with ID '{name}'")
        else:
            networks.append(network)
    if not networks:
        network_params = {'router:external': True, "status": "ACTIVE"}
        if name is not None:
            network_params['name'] = name
        networks += neutron.list_networks(**network_params)
    if not networks and name:
        raise ValueError("No such external network with name or ID "
                         f"'{name}'")
    return networks


def get_external_network_id():
    return tobiko.setup_fixture(ExternalNetworkStackFixture).network_id


def get_external_network():
    return tobiko.setup_fixture(ExternalNetworkStackFixture).network_details


def has_external_network():
    return tobiko.setup_fixture(ExternalNetworkStackFixture).has_network


skip_unless_has_external_network = tobiko.skip_unless(
    'External network not found', has_external_network)


def get_floating_network_id():
    return tobiko.setup_fixture(FloatingNetworkStackFixture).network_id


def get_floating_network():
    return tobiko.setup_fixture(FloatingNetworkStackFixture).network_details


def has_floating_network():
    return tobiko.setup_fixture(FloatingNetworkStackFixture).has_network


skip_unless_has_floating_network = tobiko.skip_unless(
    'Floating network not found', has_floating_network)


class DefaultNameserversFixture(tobiko.SharedFixture):

    remove_local_ips = True
    max_count = 3
    ip_version = None

    nameservers: typing.Optional[tobiko.Selection[netaddr.IPAddress]] = None

    @property
    def ssh_client(self):
        host = tobiko.tobiko_config().neutron.nameservers_host
        if host is None:
            return ssh.ssh_proxy_client()
        else:
            return ssh.ssh_client(host)

    @property
    def filenames(self):
        return tuple(tobiko.tobiko_config().neutron.nameservers_filenames)

    def setup_fixture(self):
        self.nameservers = self.list_nameservers()

    def list_nameservers(self) -> tobiko.Selection[netaddr.IPAddress]:
        nameservers: tobiko.Selection[netaddr.IPAddress]
        if has_external_network():
            network_id = get_external_network_id()
            nameservers = neutron.list_network_nameservers(
                network_id=network_id)
            LOG.debug("Nameservers copied from external network: "
                      f"{nameservers}")
        else:
            # Copy nameservers from target host
            nameservers = sh.list_nameservers(ssh_client=self.ssh_client,
                                              ip_version=self.ip_version,
                                              filenames=self.filenames)
            if self.remove_local_ips:
                local_ips = ip.list_ip_addresses(scope='host')
                if local_ips:
                    # Filter out all local IPs
                    nameservers = tobiko.Selection[netaddr.IPAddress](
                        nameserver for nameserver in nameservers
                        if nameserver not in local_ips)
            LOG.debug(f"Nameservers copied from host: {nameservers}")
        if self.max_count:
            # Keep only up to max_count nameservers
            actual_count = len(nameservers)
            if actual_count > self.max_count:
                LOG.waring("Limit the number of nameservers from "
                           f"{actual_count} to {self.max_count}: "
                           f"{nameservers}")
                nameservers = tobiko.Selection[netaddr.IPAddress](
                    nameservers[:self.max_count])
        return nameservers


DEFAULT_NAMESERVERS_FIXTURE = DefaultNameserversFixture


def default_nameservers(
        ip_version: typing.Optional[int] = None) -> \
        tobiko.Selection[netaddr.IPAddress]:
    nameservers = tobiko.setup_fixture(
        DEFAULT_NAMESERVERS_FIXTURE).nameservers
    if ip_version is not None:
        nameservers = nameservers.with_attributes(version=ip_version)
    return nameservers



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import socket

import tobiko
from tobiko.shell.sh import _execute


class HostnameError(tobiko.TobikoException):
    message = "Unable to get hostname from host: {error}"


def get_hostname(ssh_client=None, **execute_params):
    if ssh_client is False:
        return socket.gethostname()
    else:
        result = _execute.execute('hostname', stdin=False, stdout=True,
                                  stderr=True, expect_exit_status=None,
                                  ssh_client=ssh_client, **execute_params)
        if result.exit_status or not result.stdout:
            raise HostnameError(error=result.stderr)
        return result.stdout.splitlines()[0].strip()



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import time

import testtools


class TestCaseTest(testtools.TestCase):

    def test_with_timeout(self):

        class MyTest(testtools.TestCase):

            _testcase_timeout = 1.

            def test_busy(self):
                while True:
                    time.sleep(0.)

        test_case = MyTest('test_busy')
        test_result = testtools.TestResult()
        test_case.run(test_result)

        reported_test_case, reported_error = test_result.errors[-1]
        self.assertIs(test_case, reported_test_case)
        self.assertIn('TestCaseTimeoutError', reported_error)



# Copyright (c) 2018 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import argparse
import io
import os
import subprocess
import sys

import tobiko
from tobiko.cmd import fixture as _fixture
from tobiko.tests.unit import test_fixture
from tobiko.tests import unit


MyFixture = test_fixture.MyFixture
MyFixture2 = test_fixture.MyFixture2
MyRequiredFixture = test_fixture.MyRequiredFixture
canonical_name = test_fixture.canonical_name


class ExitCalled(Exception):
    pass


class FixtureUtilTest(unit.TobikoUnitTest):

    command_name = 'tobiko-fixture'
    command_class = _fixture.FixtureUtil

    test_path = os.path.dirname(__file__)
    top_dir = os.path.dirname(os.path.dirname(tobiko.__file__))

    required_fixture = tobiko.required_fixture(MyRequiredFixture)

    def setUp(self):
        super(FixtureUtilTest, self).setUp()
        self.mock_error = self.patch(argparse.ArgumentParser, 'error',
                                     side_effect=self.fail)

    def patch_argv(self, subcommand=None, arguments=None, filters=None,
                   **options):
        subcommand = subcommand or 'list'
        arguments = arguments or []
        if options:
            arguments += make_options(**options)
        if filters:
            arguments += list(filters)
        argv = [self.command_name, subcommand] + arguments
        return self.patch(sys, 'argv', argv)

    def test_init(self, subcommand=None, arguments=None, filters=None,
                  config_file=None, repo_type=None,
                  repo_url=None, test_path=None, top_dir=None,
                  group_regex=None, blacklist_file=None,
                  whitelist_file=None, black_regex=None):
        self.patch_argv(subcommand=subcommand, arguments=arguments)
        command = self.command_class()
        self.mock_error.assert_not_called()
        args = command.args
        self.assertIsNotNone(args)
        self.assertEqual(filters or [], args.filters)
        self.assertEqual(config_file or '.stestr.conf', args.config)
        self.assertEqual(subcommand or 'list', args.subcommand)
        self.assertEqual(repo_type or 'file', args.repo_type)
        self.assertEqual(repo_url, args.repo_url)
        self.assertEqual(test_path, args.test_path)
        self.assertEqual(top_dir, args.top_dir)
        self.assertEqual(group_regex, args.group_regex)
        self.assertEqual(blacklist_file, args.blacklist_file)
        self.assertEqual(whitelist_file, whitelist_file)
        self.assertEqual(black_regex, args.black_regex)

    def test_init_with_list(self):
        self.test_init(subcommand='list')

    def test_init_with_seutp(self):
        self.test_init(subcommand='setup')

    def test_init_with_cleanup(self):
        self.test_init(subcommand='cleanup')

    def test_init_with_c(self):
        self.test_init(arguments=['-c', 'some-config-file'],
                       config_file='some-config-file')

    def test_init_with_config(self):
        self.test_init(arguments=['--config', 'some-config-file'],
                       config_file='some-config-file')

    def test_init_with_r(self):
        self.test_init(arguments=['-r', 'file'], repo_type='file')

    def test_init_with_repo_type_file(self):
        self.test_init(arguments=['--repo-type', 'file'], repo_type='file')

    def test_init_with_repo_type_sql(self):
        self.assertRaises(self.failureException, self.test_init,
                          arguments=['--repo-type', 'sql'])

    def test_init_with_u(self):
        self.test_init(arguments=['-u', 'some-url'], repo_url='some-url')

    def test_init_with_repo_url(self):
        self.test_init(arguments=['--repo-url', 'some-url'],
                       repo_url='some-url')

    def test_init_with_filters(self):
        self.test_init(arguments=['a', 'b', 'c'], filters=['a', 'b', 'c'])

    def test_init_with_t(self):
        self.test_init(arguments=['-t', 'some/test/path'],
                       test_path='some/test/path')

    def test_init_with_test_path(self):
        self.test_init(arguments=['--test-path', 'some/test/path'],
                       test_path='some/test/path')

    def test_init_with_top_dir(self):
        self.test_init(arguments=['--top-dir', 'some/top/dir'],
                       top_dir='some/top/dir')

    def test_init_with_g(self):
        self.test_init(arguments=['-g', 'some-regex'],
                       group_regex='some-regex')

    def test_init_with_group_regex(self):
        self.test_init(arguments=['--group-regex', 'some-regex'],
                       group_regex='some-regex')

    def test_init_with_group_regex2(self):
        self.test_init(arguments=['--group_regex', 'some-regex'],
                       group_regex='some-regex')

    def test_init_with_b(self):
        self.test_init(arguments=['-b', 'some/blacklist-file'],
                       blacklist_file='some/blacklist-file')

    def test_init_with_blacklist_file(self):
        self.test_init(arguments=['--blacklist-file', 'some/blacklist-file'],
                       blacklist_file='some/blacklist-file')

    def test_init_with_w(self):
        self.test_init(arguments=['-w', 'some/whitelist-file'],
                       whitelist_file='some/whitelist-file')

    def test_init_with_whitelist_file(self):
        self.test_init(arguments=['--whitelist-file', 'some/whitelist-file'],
                       whitelist_file='some/whitelist-file')

    def test_init_with_B(self):
        self.test_init(arguments=['-B', 'some-black-regex'],
                       black_regex='some-black-regex')

    def test_init_with_blacklist_regex(self):
        self.test_init(arguments=['--black-regex', 'some-black-regex'],
                       black_regex='some-black-regex')

    def test_main(self, subcommand=None, config_file=None, repo_type=None,
                  repo_url=None, test_path=None, top_dir=None,
                  group_regex=None, blacklist_file=None, whitelist_file=None,
                  black_regex=None, filters=None):
        test_path = test_path or self.test_path
        top_dir = top_dir or self.top_dir
        self.setup_file_repo(top_dir=top_dir)
        self.patch_argv(subcommand=subcommand, config_file=config_file,
                        repo_type=repo_type, repo_url=repo_url,
                        test_path=test_path, top_dir=top_dir,
                        group_regex=group_regex,
                        blacklist_file=blacklist_file,
                        whitelist_file=whitelist_file,
                        black_regex=black_regex, filters=filters)
        stdout = self.patch(sys, 'stdout', io.StringIO())
        _fixture.main()
        self.mock_error.assert_not_called()
        return stdout

    def test_list(self, fixture1=MyFixture, fixture2=MyFixture2):
        stdout = self.test_main(subcommand='list')
        written_lines = stdout.getvalue().splitlines()
        self.assertIn(canonical_name(fixture1), written_lines)
        self.assertIn(canonical_name(fixture2), written_lines)
        self.assertIn(canonical_name(MyRequiredFixture),
                      written_lines)

    def test_list_with_filters(self, fixture=MyFixture):
        stdout = self.test_main(subcommand='list', filters=[self.id()])
        written_lines = stdout.getvalue().splitlines()
        self.assertEqual([canonical_name(fixture),
                          canonical_name(MyRequiredFixture)],
                         written_lines)
        self.assertIn(canonical_name(MyRequiredFixture),
                      written_lines)

    def test_setup(self, fixture=MyFixture, fixture2=MyFixture2):
        stdout = self.test_main(subcommand='setup')
        written_lines = stdout.getvalue().splitlines()
        for obj in [fixture, fixture2, MyRequiredFixture]:
            self.assertIn(canonical_name(obj), written_lines)
            tobiko.get_fixture(obj).setup_fixture.assert_called_once_with()
            tobiko.get_fixture(obj).cleanup_fixture.assert_not_called()

    def test_setup_with_filters(self, fixture=MyFixture):
        stdout = self.test_main(subcommand='setup', filters=[self.id()])
        written_lines = stdout.getvalue().splitlines()
        self.assertEqual([canonical_name(fixture),
                          canonical_name(MyRequiredFixture)],
                         written_lines)
        for obj in [fixture, MyRequiredFixture]:
            tobiko.get_fixture(obj).setup_fixture.assert_called_once_with()
            tobiko.get_fixture(obj).cleanup_fixture.assert_not_called()

    def test_cleanup(self, fixture=MyFixture, fixture2=MyFixture2):
        stdout = self.test_main(subcommand='cleanup')
        written_lines = stdout.getvalue().splitlines()
        for obj in [fixture, fixture2, MyRequiredFixture]:
            self.assertIn(canonical_name(obj), written_lines)
            tobiko.get_fixture(obj).setup_fixture.assert_not_called()
            tobiko.get_fixture(obj).cleanup_fixture.assert_called_once_with()

    def test_cleanup_with_filters(self, fixture=MyFixture):
        stdout = self.test_main(subcommand='cleanup', filters=[self.id()])
        written_lines = stdout.getvalue().splitlines()
        self.assertEqual([canonical_name(fixture),
                          canonical_name(MyRequiredFixture)],
                         written_lines)
        for obj in [fixture, MyRequiredFixture]:
            tobiko.get_fixture(obj).setup_fixture.assert_not_called()
            tobiko.get_fixture(obj).cleanup_fixture.assert_called_once_with()

    def setup_file_repo(self, top_dir):
        if not os.path.isdir(os.path.join(top_dir, '.stestr')):
            command = ['stestr', '--top-dir', top_dir, 'init']
            subprocess.check_call(command)


def make_options(config_file=None, repo_type=None, repo_url=None,
                 test_path=None, top_dir=None, group_regex=None,
                 blacklist_file=None, whitelist_file=None,
                 black_regex=None):
    options = []
    if config_file:
        options += ['--config', config_file]
    if repo_type:
        options += ['--repo-type', repo_type]
    if repo_url:
        options += ['--repo-url', repo_url]
    if test_path:
        options += ['--test-path', test_path]
    if top_dir:
        options += ['--top-dir', top_dir]
    if group_regex:
        options += ['--group-regex', group_regex]
    if blacklist_file:
        options += ['--blacklist-file', blacklist_file]
    if whitelist_file:
        options += ['--whitelist-file', whitelist_file]
    if black_regex:
        options += ['--black-regex', black_regex]
    return options



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import mock

from tobiko import podman
from tobiko.tests import unit


class TestShell(unit.TobikoUnitTest):

    @mock.patch('tobiko.shell.sh.execute')
    def test_discover_podman_socket(self, mock_execute):
        class FakeProcess:
            exit_status = 0
            stdout = '/run/podman/io.podman'
            stderr = ''
        mock_execute.return_value = FakeProcess()
        self.assertEqual(
            podman.discover_podman_socket(),
            '/run/podman/io.podman'
        )

    @mock.patch('tobiko.shell.sh.execute')
    def test_discover_podman_socket_none_result(self, mock_execute):
        class FakeProcess:
            exit_status = 1
            stdout = ''
            stderr = 'boom'
        mock_execute.return_value = FakeProcess()
        self.assertRaises(
            podman.PodmanSocketNotFoundError,
            podman.discover_podman_socket
        )

    @mock.patch('tobiko.shell.sh.execute')
    def test_discover_podman_socket_with_exit_code(self, mock_execute):
        class FakeProcess:
            exit_status = 0
            stdout = ''
            stderr = 'boom'
        mock_execute.return_value = FakeProcess()
        self.assertRaises(
            podman.PodmanSocketNotFoundError,
            podman.discover_podman_socket
        )

    @mock.patch('tobiko.shell.sh.execute')
    def test_is_podman_running(self, mock_execute):
        class FakeProcess:
            exit_status = 0
            stdout = '/run/podman/io.podman'
            stderr = ''
        mock_execute.return_value = FakeProcess()
        self.assertEqual(podman.is_podman_running(), True)

    @mock.patch('tobiko.shell.sh.execute')
    def test_is_podman_running_without_socket(self, mock_execute):
        class FakeProcess:
            exit_status = 1
            stdout = ''
            stderr = 'boom'
        mock_execute.return_value = FakeProcess()
        self.assertEqual(podman.is_podman_running(), False)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing  # noqa

import tobiko
from tobiko.openstack.neutron import _client


# Agent binary names
DHCP_AGENT = 'neutron-dhcp-agent'
L3_AGENT = 'neutron-l3-agent'
METADATA_AGENT = 'neutron-metadata-agent'
OPENVSWITCH_AGENT = 'neutron-openvswitch-agent'
OVN_CONTROLLER = 'ovn-controller'
OVN_METADATA_AGENT = 'networking-ovn-metadata-agent'


class AgentNotFoundOnHost(tobiko.TobikoException):
    message = ("Agent {agent_type!s} not found on the host {host!s}")


class NetworkingAgentFixture(tobiko.SharedFixture):

    agents = None

    def setup_fixture(self):
        self.agents = _client.list_agents()


def list_networking_agents(**attributes):
    return tobiko.setup_fixture(
        NetworkingAgentFixture).agents.with_items(**attributes)


def missing_networking_agents(count=1, **params):
    agents = list_networking_agents(**params)
    return max(0, count - len(agents))


def has_networking_agents(count=1, **params):
    return not missing_networking_agents(count=count, **params)


DecoratorType = typing.Callable[[typing.Union[typing.Callable, typing.Type]],
                                typing.Union[typing.Callable, typing.Type]]


def skip_if_missing_networking_agents(binary: typing.Optional[str] = None,
                                      count: int = 1, **params) -> \
        DecoratorType:
    if binary is not None:
        params['binary'] = binary
    message = "missing {return_value!r} agent(s)"
    if params:
        message += " with {!s}".format(
            ', '.join("{!s}={!r}".format(k, v) for k, v in params.items()))
    return tobiko.skip_if(message, missing_networking_agents, count=count,
                          **params)


def skip_unless_is_ovn():
    '''Skip the test if OVN is not configured'''
    from tobiko.tripleo import overcloud
    from tobiko.tripleo import containers
    if overcloud.has_overcloud():
        message = "OVN is not configured"
        return tobiko.skip_unless(message, containers.ovn_used_on_overcloud)
    else:
        return skip_if_missing_networking_agents(OVN_CONTROLLER)


def skip_unless_is_ovs():
    '''Skip the test if openvswitch agent does not exist'''
    return skip_if_missing_networking_agents(OPENVSWITCH_AGENT)



from __future__ import absolute_import

import testtools


from tobiko.tests.faults.ha import cloud_disruptions
from tobiko.tripleo import pacemaker
from tobiko.tripleo import undercloud


@undercloud.skip_if_missing_undercloud
class DisruptTripleoInstanceHaTest(testtools.TestCase):

    @pacemaker.skip_if_instanceha_not_delpoyed
    def test_instanceha_evacuation_hard_reset(self):
        cloud_disruptions.check_iha_evacuation_hard_reset()

    @pacemaker.skip_if_instanceha_not_delpoyed
    def test_instanceha_evacuation_network_disruption(self):
        cloud_disruptions.check_iha_evacuation_network_disruption()

    @pacemaker.skip_if_instanceha_not_delpoyed
    def test_instanceha_evacuation_hard_reset_shutoff_instance(self):
        cloud_disruptions.check_iha_evacuation_hard_reset_shutoff_instance()

    # @pacemaker.skip_if_instanceha_not_delpoyed
    # def test_check_instanceha_evacuation_evac_image_vm(self):
    #     overcloud_health_checks(passive_checks_only=True)
    #     cloud_disruptions.check_iha_evacuation_evac_image_vm()



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import time as _time
import typing

from tobiko.common import _exception


Seconds = typing.Optional[float]


class SecondsValueError(_exception.TobikoException):
    message = "Invalid seconds value: {seconds}"


ToSecondsValue = typing.Union[float, int, str, bytearray, None]


def to_seconds(value: ToSecondsValue) -> Seconds:
    if value is None:
        return None
    else:
        return to_seconds_float(value)


def to_seconds_float(value: ToSecondsValue) -> float:
    return value and max(0., float(value)) or 0.


def time() -> float:
    return _time.time()


def sleep(seconds: Seconds):
    _time.sleep(to_seconds_float(seconds))


def true_seconds(*seconds: Seconds) -> typing.List[float]:
    result = []
    for value in seconds:
        value = to_seconds(value)
        if value is not None:
            result.append(value)
    return result


def min_seconds(*seconds: Seconds) -> Seconds:
    values = true_seconds(*seconds)
    if values:
        return min(values)
    else:
        return None


def max_seconds(*seconds: Seconds) -> Seconds:
    values = true_seconds(*seconds)
    if values:
        return max(values)
    else:
        return None



#!/usr/bin/env python3
# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import psutil
import signal
import sys
import subprocess


TOP_DIR = os.path.dirname(os.path.dirname(__file__))
if TOP_DIR not in sys.path:
    sys.path.insert(0, TOP_DIR)

from tools import common  # noqa

LOG = common.get_logger(__name__)


# Output dirs
TOX_REPORT_DIR = common.normalize_path(
    os.environ.get('TOX_REPORT_DIR', os.getcwd()))

TOX_REPORT_NAME = os.environ.get('TOX_REPORT_NAME', 'test_results')
TOX_REPORT_PREFIX = os.path.join(TOX_REPORT_DIR, TOX_REPORT_NAME)

TOX_REPORT_LOG = os.environ.get(
    'TOX_REPORT_LOG', TOX_REPORT_PREFIX + '.log')

TOX_REPORT_SUBUNIT = os.environ.get(
    'TOX_REPORT_SUBUNIT', TOX_REPORT_PREFIX + '.subunit')

TOX_REPORT_HTML = os.environ.get(
    'TOX_REPORT_HTML', TOX_REPORT_PREFIX + '.html')

TOX_REPORT_XML = os.environ.get(
    'TOX_REPORT_XML', TOX_REPORT_PREFIX + '.xml')

TOX_RUN_TESTS_TIMEOUT = float(os.environ.get('TOX_RUN_TESTS_TIMEOUT') or 0.)

TOX_PYDEV_DEBUG = bool(
    os.environ.get('TOX_PYDEV_DEBUG', 'false').lower() in
    ['true', 'yes', '1'])


def main():
    common.setup_logging()
    try:
        succeeded = run_tests()
        if succeeded:
            LOG.info('SUCCEEDED')
            sys.exit(0)
        else:
            LOG.info('FAILED')
            sys.exit(1)

    except Exception:
        LOG.exception('ERROR')
        sys.exit(2)


def run_tests():
    setup_timeout()
    cleanup_report_dir()
    log_environ()

    if TOX_PYDEV_DEBUG:
        debug_test_cases()
        return True

    succeeded = True
    try:
        run_test_cases()
    except (subprocess.CalledProcessError,
            subprocess.TimeoutExpired,
            ProcessLookupError) as ex:
        LOG.error(f"Error while running test cases.\n{ex}")
        succeeded = False

    try:
        log_tests_results()
    except subprocess.CalledProcessError:
        if succeeded:
            raise

    make_subunit_file()
    make_html_file()
    try:
        make_xml_file()
    except subprocess.CalledProcessError:
        if succeeded:
            raise

    return succeeded


def setup_timeout():

    if TOX_RUN_TESTS_TIMEOUT > 0.:

        def handle_timeout(_signum, _frame):
            LOG.error(
                f"run_tests.py timeout out after {TOX_RUN_TESTS_TIMEOUT} "
                "seconds")
            terminate_childs()
            raise subprocess.TimeoutExpired("run_tests.py",
                                            TOX_RUN_TESTS_TIMEOUT)

        signal.setitimer(signal.ITIMER_REAL, TOX_RUN_TESTS_TIMEOUT)
        signal.signal(signal.SIGALRM, handle_timeout)
        LOG.debug(f'Run tests timeout set as {TOX_RUN_TESTS_TIMEOUT} seconds')


def terminate_childs():
    current_process = psutil.Process()
    children = current_process.children(recursive=False)
    for child in children:
        LOG.debug(f"Interrupt child process execution (pid={child.pid})")
        os.kill(child.pid, signal.SIGINT)
    for child in children:
        LOG.debug("Wait for top-child process termination "
                  f"(pid={child.pid})...")
        os.waitpid(child.pid, 0)


def cleanup_report_dir():
    for report_file in [TOX_REPORT_LOG, TOX_REPORT_SUBUNIT, TOX_REPORT_HTML,
                        TOX_REPORT_XML]:
        if not common.make_dir(os.path.dirname(report_file)):
            common.remove_file(report_file)


def log_environ():
    common.execute('env | sort >> "{log_file}"', log_file=TOX_REPORT_LOG,
                   capture_stdout=False)


def log_tests_results():
    common.execute('stestr last --all-attachments >> "{log_file}"',
                   log_file=TOX_REPORT_LOG,
                   capture_stdout=False,
                   check=False)


def debug_test_cases():
    common.execute_python('-m testtools.run {posargs}',
                          posargs=common.get_posargs(),
                          capture_stdout=False)


def run_test_cases():
    common.execute('stestr run {posargs}',
                   posargs=common.get_posargs(),
                   capture_stdout=False)


def make_subunit_file():
    common.execute('stestr last --subunit > "{subunit_file}"',
                   subunit_file=TOX_REPORT_SUBUNIT,
                   capture_stdout=False)


def make_html_file():
    common.execute('subunit2html "{subunit_file}" "{html_file}"',
                   subunit_file=TOX_REPORT_SUBUNIT,
                   html_file=TOX_REPORT_HTML,
                   capture_stdout=False)


def make_xml_file():
    common.execute('subunit2junitxml "{subunit_file}" -o "{xml_file}"',
                   subunit_file=TOX_REPORT_SUBUNIT,
                   xml_file=TOX_REPORT_XML,
                   capture_stdout=False)


if __name__ == '__main__':
    main()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.neutron import _agent
from tobiko.openstack.neutron import _client
from tobiko.openstack.neutron import _cidr
from tobiko.openstack.neutron import _extension
from tobiko.openstack.neutron import _port
from tobiko.openstack.neutron import _network


SERVER = 'neutron-server'
DHCP_AGENT = _agent.DHCP_AGENT
L3_AGENT = _agent.L3_AGENT
METADATA_AGENT = _agent.METADATA_AGENT
OPENVSWITCH_AGENT = _agent.OPENVSWITCH_AGENT
OVN_CONTROLLER = _agent.OVN_CONTROLLER
OVN_METADATA_AGENT = _agent.OVN_METADATA_AGENT
AgentNotFoundOnHost = _agent.AgentNotFoundOnHost
skip_if_missing_networking_agents = _agent.skip_if_missing_networking_agents
skip_unless_is_ovn = _agent.skip_unless_is_ovn
skip_unless_is_ovs = _agent.skip_unless_is_ovs
list_networking_agents = _agent.list_networking_agents

NeutronClientFixture = _client.NeutronClientFixture
ServiceUnavailable = _client.ServiceUnavailable
neutron_client = _client.neutron_client
get_neutron_client = _client.get_neutron_client
find_subnet = _client.find_subnet
find_port = _client.find_port
list_ports = _client.list_ports
list_subnets = _client.list_subnets
list_subnet_cidrs = _client.list_subnet_cidrs
list_agents = _client.list_agents
get_floating_ip = _client.get_floating_ip
get_router = _client.get_router
get_port = _client.get_port
get_subnet = _client.get_subnet
list_l3_agent_hosting_routers = _client.list_l3_agent_hosting_routers
find_l3_agent_hosting_router = _client.find_l3_agent_hosting_router
list_dhcp_agent_hosting_network = _client.list_dhcp_agent_hosting_network

NoSuchPort = _client.NoSuchPort
NoSuchRouter = _client.NoSuchRouter
NoSuchSubnet = _client.NoSuchSubnet

new_ipv4_cidr = _cidr.new_ipv4_cidr
new_ipv6_cidr = _cidr.new_ipv6_cidr

get_networking_extensions = _extension.get_networking_extensions
missing_networking_extensions = _extension.missing_networking_extensions
has_networking_extensions = _extension.has_networking_extensions

skip_if_missing_networking_extensions = (
    _extension.skip_if_missing_networking_extensions)

find_port_ip_address = _port.find_port_ip_address
list_port_ip_addresses = _port.list_port_ip_addresses
find_device_ip_address = _port.find_device_ip_address
list_device_ip_addresses = _port.list_device_ip_addresses

NeutronNetworkFixture = _network.NeutronNetworkFixture
NoSuchNetwork = _network.NoSuchNetwork
create_network = _network.create_network
delete_network = _network.delete_network
get_network = _network.get_network
find_network = _network.find_network
list_networks = _network.list_networks
list_network_nameservers = _network.list_network_nameservers



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.shell.sh import _execute


class UptimeError(tobiko.TobikoException):
    message = "Unable to get uptime from host: {error}"


def get_uptime(**execute_params):
    """Returns the number of seconds passed since last host reboot

    It reads and parses remote special file /proc/uptime and returns a floating
    point value that represents the number of seconds passed since last host
    reboot
    """
    result = _execute.execute('cat /proc/uptime', stdin=False, stdout=True,
                              stderr=True, expect_exit_status=None,
                              **execute_params)
    output = result.stdout and result.stdout.strip()
    if result.exit_status or not output:
        raise UptimeError(error=result.stderr)

    uptime_line = output.splitlines()[0]
    uptime_string = uptime_line.split()[0]
    return float(uptime_string)



types = """
type ListPodData (
    id: string,
    name: string,
    createdat: string,
    cgroup: string,
    status: string,
    labels: [string]string,
    numberofcontainers: string,
    containersinfo: []ListPodContainerInfo
)
type ListPodContainerInfo (
    name: string,
    id: string,
    status: string
)
"""


class ServicePod:

    def StartPod(self, name: str) -> str:
        """return pod"""
        return {  # type: ignore
            "pod": "135d71b9495f7c3967f536edad57750bfd"
                   "b569336cd107d8aabab45565ffcfb6",
            "name": name
        }

    def GetPod(self, name: str) -> str:
        """return pod: ListPodData"""
        return {  # type: ignore
            "pod": {
                "cgroup": "machine.slice",
                "containersinfo": [
                  {
                    "id": "1840835294cf076a822e4e12ba4152411f131"
                          "bd869e7f6a4e8b16df9b0ea5c7f",
                    "name": "1840835294cf-infra",
                    "status": "running"
                  },
                  {
                    "id": "49a5cce72093a5ca47c6de86f10ad7bb36391e2"
                          "d89cef765f807e460865a0ec6",
                    "name": "upbeat_murdock",
                    "status": "running"
                  }
                ],
                "createdat": "2018-12-07 13:10:15.014139258 -0600 CST",
                "id": "135d71b9495f7c3967f536edad57750bfdb569336cd"
                      "107d8aabab45565ffcfb6",
                "name": name,
                "numberofcontainers": "2",
                "status": "Running"
            }
        }

    def GetVersion(self) -> str:
        """return version"""
        return {"version": "testing"}  # type: ignore



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr

import tobiko
from tobiko.openstack.neutron import _client


class NoSuchNetwork(tobiko.ObjectNotFound):
    message = "No such network found for {id!r}"


NeutronNetworkType = typing.Dict[str, typing.Any]


def get_network(network, client=None, **params) -> NeutronNetworkType:
    try:
        return _client.neutron_client(
            client).show_network(network, **params)['network']
    except _client.neutronclient.exceptions.NotFound as ex:
        raise NoSuchNetwork(id=network) from ex


def list_networks(client=None, **params) -> \
        tobiko.Selection[NeutronNetworkType]:
    networks = _client.neutron_client(client).list_networks(
        **params)['networks']
    return tobiko.select(networks)


_RAISE_ERROR = object()


def find_network(client=None, unique=False, default=_RAISE_ERROR,
                 **attributes) -> NeutronNetworkType:
    """Look for a network matching some property values"""
    networks = list_networks(client=client, **attributes)
    if default is _RAISE_ERROR or networks:
        if unique:
            return networks.unique
        else:
            return networks.first
    else:
        return default


def create_network(client=None, **params) -> NeutronNetworkType:
    return _client.neutron_client(client).create_network(
        body={'network': params})['network']


def delete_network(network, client=None):
    return _client.neutron_client(client).delete_network(network=network)


class NeutronNetworkFixture(_client.HasNeutronClientFixture):

    details: typing.Optional[NeutronNetworkType] = None

    def __init__(self, name: typing.Optional[str] = None,
                 obj: _client.NeutronClientType = None):
        super(NeutronNetworkFixture, self).__init__(obj=obj)
        if name is None:
            name = self.fixture_name
        self.name: str = name

    @property
    def id(self):
        return self.details['id']

    def setup_fixture(self):
        super(NeutronNetworkFixture, self).setup_fixture()
        self.name = self.fixture_name
        self.details = create_network(client=self.client, name=self.name)
        self.addCleanup(delete_network, network=self.id, client=self.client)


def list_network_nameservers(network_id: typing.Optional[str] = None,
                             ip_version: typing.Optional[int] = None) -> \
        tobiko.Selection[netaddr.IPAddress]:
    subnets = _client.list_subnets(network_id=network_id)
    nameservers = tobiko.Selection[netaddr.IPAddress](
        netaddr.IPAddress(nameserver)
        for subnets in subnets
        for nameserver in subnets['dns_nameservers'])
    if ip_version is not None:
        nameservers = nameservers.with_attributes(version=ip_version)
    return nameservers



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing  # noqa

import netaddr
import testtools

import tobiko
from tobiko import config
from tobiko.openstack import stacks
from tobiko.shell import ip
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.tests.functional.shell import _fixtures


CONF = config.CONF

SshClientType = typing.Union[bool, None, ssh.SSHClientFixture]


class PingTest(testtools.TestCase):

    ssh_client: SshClientType = False

    @property
    def execute_params(self):
        return dict(ssh_client=self.ssh_client)

    def test_ping_reachable_address(self):
        result = ping.ping('127.0.0.1', count=3,
                           **self.execute_params)
        self.assertIsNone(result.source)
        self.assertEqual(netaddr.IPAddress('127.0.0.1'), result.destination)
        result.assert_transmitted()
        result.assert_replied()

    def test_ping_reachable_hostname(self):
        result = ping.ping('localhost', count=3, **self.execute_params)
        self.assertIsNone(result.source)
        self.assertIsNotNone(result.destination)
        result.assert_transmitted()
        result.assert_replied()

    def test_ping_unreachable_address(self):
        result = ping.ping('1.2.3.4', count=3, check=False,
                           **self.execute_params)
        self.assertIsNone(result.source)
        self.assertIn(result.destination, [netaddr.IPAddress('1.2.3.4'),
                                           None])
        if result.destination is not None:
            result.assert_transmitted()
        result.assert_not_replied()

    def test_ping_invalid_ip(self):
        try:
            result = ping.ping('0.1.2.3', count=1,
                               **self.execute_params)
        except ping.PingError as ex:
            self.assertIn(ex, [
                ping.ConnectPingError(details='Invalid argument'),
                ping.ConnectPingError(details='Network is unreachable'),
                ping.SendToPingError(details='No route to host'),
            ])
        else:
            self.assertIsNone(result.source)
            self.assertEqual(netaddr.IPAddress('0.1.2.3'),
                             result.destination)
            result.assert_transmitted()
            result.assert_not_replied()

    def test_ping_unreachable_hostname(self):
        ex = self.assertRaises(ping.PingError, ping.ping,
                               'unreachable-host', count=3,
                               **self.execute_params)
        self.assertIn(ex, [
            ping.UnknowHostError(details='unreachable-host'),
            ping.UnknowHostError(
                details='Temporary failure in name resolution'),
            ping.BadAddressPingError(address='unreachable-host'),
            ping.UnknowHostError(
                details='Name or service not known')
        ])

    def test_ping_until_received(self):
        result = ping.ping_until_received('127.0.0.1', count=3,
                                          **self.execute_params)
        self.assertIsNone(result.source)
        self.assertEqual(netaddr.IPAddress('127.0.0.1'), result.destination)
        result.assert_transmitted()
        result.assert_replied()

    def test_ping_until_received_unreachable(self):
        ex = self.assertRaises(ping.PingError, ping.ping_until_received,
                               '1.2.3.4', count=3, timeout=6,
                               **self.execute_params)
        self.assertIn(ex, [
            ping.PingFailed(timeout=6, count=0, expected_count=3,
                            message_type='received'),
            ping.ConnectPingError(details='Network is unreachable')])

    def test_ping_until_unreceived_reachable(self):
        ex = self.assertRaises(ping.PingFailed, ping.ping_until_unreceived,
                               '127.0.0.1', count=3, timeout=6,
                               **self.execute_params)
        self.assertEqual(6, ex.timeout)
        self.assertEqual(0, ex.count)
        self.assertEqual(3, ex.expected_count)
        self.assertEqual('unreceived', ex.message_type)

    def test_ping_until_unreceived_unreachable(self):
        result = ping.ping_until_unreceived('1.2.3.4', count=3, check=False,
                                            **self.execute_params)
        self.assertIsNone(result.source)
        if result.destination is None:
            result.assert_not_transmitted()
        else:
            self.assertEqual(result.destination, netaddr.IPAddress('1.2.3.4'))
            result.assert_transmitted()
        result.assert_not_replied()

    def test_ping_reachable_with_timeout(self):
        ex = self.assertRaises(ping.PingFailed, ping.ping, '127.0.0.1',
                               count=20, timeout=1.,
                               **self.execute_params)
        self.assertEqual(1., ex.timeout)
        self.assertEqual(20, ex.expected_count)
        self.assertEqual('transmitted', ex.message_type)

    def test_ping_hosts(self):
        try:
            sh.execute('[ -x /sbin/ip ]', ssh_client=self.ssh_client)
        except sh.ShellCommandFailed:
            self.skipTest("'/sbin/ip' command not found")
        ips = ip.list_ip_addresses(**self.execute_params)
        reachable_ips, unreachable_ips = ping.ping_hosts(
            ips, **self.execute_params)

        expected_reachable = [i for i in ips if i in reachable_ips]
        self.assertEqual(expected_reachable, reachable_ips)
        expected_unreachable = [i for i in ips if i not in reachable_ips]
        self.assertEqual(expected_unreachable, unreachable_ips)


@ssh.skip_unless_has_ssh_proxy_jump
class ProxyPingTest(PingTest):
    ssh_client = None


class NamespacePingTest(PingTest):

    namespace = tobiko.required_setup_fixture(
        _fixtures.NetworkNamespaceFixture)

    @property
    def ssh_client(self):
        return self.namespace.ssh_client

    @property
    def network_namespace(self):
        return self.namespace.network_namespace

    @property
    def execute_params(self):
        return dict(ssh_client=self.ssh_client,
                    network_namespace=self.network_namespace)


class CirrosPingTest(PingTest):

    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    @property
    def ssh_client(self):
        return self.stack.ssh_client


class CentosPingTest(CirrosPingTest):

    stack = tobiko.required_setup_fixture(stacks.CentosServerStackFixture)


class UbuntuPingTest(CirrosPingTest):

    stack = tobiko.required_setup_fixture(stacks.UbuntuServerStackFixture)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import six

from tobiko import config
from tobiko.tests import unit


CONF = config.CONF
TIPLEO_CONF = CONF.tobiko.tripleo


class TripleoConfigTest(unit.TobikoUnitTest):

    def test_ssh_key_filename(self):
        self.assertIsInstance(TIPLEO_CONF.undercloud_ssh_key_filename,
                              six.string_types)


class UndercloudConfigTest(unit.TobikoUnitTest):

    def test_undercloud_ssh_hostname(self):
        value = TIPLEO_CONF.undercloud_ssh_hostname
        if value is not None:
            self.assertIsInstance(value, six.string_types)

    def test_undercloud_ssh_port(self):
        value = TIPLEO_CONF.undercloud_ssh_port
        if value is not None:
            self.assertIsInstance(value, int)
            self.assertIn(value, six.moves.range(1, 2 ** 16))

    def test_undercloud_ssh_username(self):
        self.assertIsInstance(TIPLEO_CONF.undercloud_ssh_username,
                              six.string_types)

    def test_undercloud_rcfile(self):
        for rcfile in TIPLEO_CONF.undercloud_rcfile:
            self.assertIsInstance(rcfile, six.string_types)


class OvercloudConfigTest(unit.TobikoUnitTest):

    def test_overcloud_ssh_port(self):
        value = TIPLEO_CONF.overcloud_ssh_port
        if value is not None:
            self.assertIsInstance(value, int)
            self.assertIn(value, six.moves.range(1, 2 ** 16))

    def test_overcloud_ssh_username(self):
        self.assertIsInstance(TIPLEO_CONF.overcloud_ssh_username,
                              six.string_types)

    def test_overcloud_rcfile(self):
        for rcfile in TIPLEO_CONF.overcloud_rcfile:
            self.assertIsInstance(rcfile, six.string_types)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import podman

import tobiko
from tobiko.podman import _exception
from tobiko.podman import _shell
from tobiko.shell import ssh
from tobiko.shell import sh


def get_podman_client(ssh_client=None):
    return PodmanClientFixture(ssh_client=ssh_client)


def list_podman_containers(client=None, **kwargs):
    try:
        containers = podman_client(client).containers.list(**kwargs)
    except _exception.PodmanSocketNotFoundError:
        return tobiko.Selection()
    else:
        return tobiko.select(containers)


def podman_client(obj=None):
    if obj is None:
        obj = get_podman_client()
    if tobiko.is_fixture(obj):
        obj = tobiko.setup_fixture(obj).client
    if isinstance(obj, podman.Client):
        return obj
    raise TypeError('Cannot obtain a Podman client from {!r}'.format(obj))


class PodmanClientFixture(tobiko.SharedFixture):

    client = None
    ssh_client = None

    def __init__(self, ssh_client=None):
        super(PodmanClientFixture, self).__init__()
        if ssh_client:
            self.ssh_client = ssh_client

    def setup_fixture(self):
        self.setup_ssh_client()
        self.setup_client()

    def setup_ssh_client(self):
        ssh_client = self.ssh_client
        if ssh_client is None:
            self.ssh_client = ssh_client = ssh.ssh_proxy_client() or False
            if ssh_client:
                tobiko.setup_fixture(ssh_client)
        return ssh_client

    def setup_client(self):
        # setup podman access via varlink
        podman_client_setup_cmds = \
            "sudo test -f /var/varlink_client_access_setup ||  \
            (sudo groupadd -f podman &&  \
            sudo usermod -a -G podman heat-admin && \
            sudo chmod -R o=wxr /etc/tmpfiles.d && \
            sudo echo 'd /run/podman 0770 root heat-admin' >  \
            /etc/tmpfiles.d/podman.conf && \
            sudo cp /lib/systemd/system/io.podman.socket \
            /etc/systemd/system/io.podman.socket && \
            sudo crudini --set /etc/systemd/system/io.podman.socket Socket  \
            SocketMode 0660 && \
            sudo crudini --set /etc/systemd/system/io.podman.socket Socket  \
            SocketGroup podman && \
            sudo systemctl daemon-reload && \
            sudo systemd-tmpfiles --create && \
            sudo systemctl enable --now io.podman.socket && \
            sudo chmod 777 /run/podman && \
            sudo chown -R root: /run/podman && \
            sudo chmod g+rw /run/podman/io.podman && \
            sudo chmod 777 /run/podman/io.podman && \
            sudo setenforce 0 && \
            sudo systemctl start io.podman.socket && \
            sudo touch /var/varlink_client_access_setup)"

        sh.execute(podman_client_setup_cmds, ssh_client=self.ssh_client)

        client = self.client
        if client is None:
            self.client = client = self.create_client()
        return client

    def create_client(self):
        for _ in range(360):

            try:
                podman_remote_socket = self.discover_podman_socket()
                username = self.ssh_client.connect_parameters['username']
                host = self.ssh_client.connect_parameters["hostname"]
                socket = podman_remote_socket
                podman_remote_socket_uri = \
                    'unix:/tmp/podman.sock_{}'.format(host)

                remote_uri = 'ssh://{username}@{host}{socket}'.format(
                    username=username,
                    host=host,
                    socket=socket)

                client = podman.Client(uri=podman_remote_socket_uri,
                                       remote_uri=remote_uri,
                                       identity_file='~/.ssh/id_rsa')
                client.system.ping()
                return client
            except (ConnectionRefusedError, ConnectionResetError):
                # retry
                self.create_client()

    def connect(self):
        return tobiko.setup_fixture(self).client

    def discover_podman_socket(self):
        return _shell.discover_podman_socket(ssh_client=self.ssh_client)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import signal
import traceback
import typing  # noqa

from tobiko.common import _exception
from tobiko.common import _fixture
from tobiko.common import _time


ITIMER_SIGNALS = {
    signal.SIGALRM: signal.ITIMER_REAL,
    signal.SIGVTALRM: signal.ITIMER_VIRTUAL,
    signal.SIGPROF: signal.ITIMER_PROF
}


class ITimerExpired(_exception.TobikoException):
    message = ("ITimer expired (signal_number={signal_number}):\n"
               "{stack}")


ITimerHandler = typing.Callable[[int, typing.Any], typing.Any]


class ITimer(_fixture.SharedFixture):

    signal_number: int = signal.SIGALRM
    delay: float = 0.
    interval: float = 0.

    original_delay: _time.Seconds = None
    original_interval: _time.Seconds = None
    original_handler: typing.Union[typing.Callable, int, None] = None

    def __init__(self,
                 signal_number: typing.Optional[int] = None,
                 delay: _time.Seconds = None,
                 interval: _time.Seconds = None,
                 on_timeout: typing.Optional[ITimerHandler] = None):
        super(ITimer, self).__init__()
        if signal_number is not None:
            self.signal_number = signal_number
        if delay is not None:
            self.delay = delay
        if interval is not None:
            self.interval = interval
        if on_timeout:
            setattr(self, 'on_timeout', on_timeout)

    def setup_fixture(self):
        self.setup_handler()
        self.setup_timer()

    def setup_handler(self):
        self.original_handler = signal.getsignal(self.signal_number)
        signal.signal(self.signal_number, self.on_timeout)
        self.addCleanup(self.cleanup_handler)

    def cleanup_handler(self):
        if self.original_handler is not None:
            signal.signal(self.signal_number, self.original_handler)
            del self.original_handler

    def setup_timer(self):
        self.original_delay, self.original_interval = signal.setitimer(
            self.which, self.delay, self.interval)
        self.addCleanup(self.cleanup_timer)

    def cleanup_timer(self):
        if (self.original_delay is not None and
                self.original_interval is not None):
            signal.setitimer(self.which, self.original_delay,
                             self.original_interval)
            del self.original_delay
            del self.original_interval

    @property
    def which(self):
        return ITIMER_SIGNALS[self.signal_number]

    def on_timeout(self, signal_number: int, frame: typing.Any):
        assert self.signal_number == signal_number
        stack = ''.join(traceback.format_stack(frame))
        raise ITimerExpired(signal_number=signal_number, stack=stack)


def itimer(signal_number: typing.Optional[int] = None,
           delay: _time.Seconds = None,
           interval: _time.Seconds = None,
           on_timeout: typing.Optional[ITimerHandler] = None):
    return ITimer(signal_number=signal_number,
                  delay=delay,
                  interval=interval,
                  on_timeout=on_timeout)



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import logging
import os
import sys
import traceback
import typing  # noqa

from oslo_log import log
import testtools

from tobiko.common import _config
from tobiko.common import _exception
from tobiko.common import _itimer
from tobiko.common import _logging
from tobiko.common import _time


LOG = log.getLogger(__name__)

os.environ.setdefault('PYTHON', sys.executable)


class TestCasesFinder(object):

    def __init__(self, config=None, repo_type=None, repo_url=None,
                 test_path=None, top_dir=None, group_regex=None,
                 blacklist_file=None, whitelist_file=None, black_regex=None,
                 filters=None):
        """
        :param str config: The path to the stestr config file. Must be a
            string.
        :param str repo_type: This is the type of repository to use. Valid
            choices are 'file' and 'sql'.
        :param str repo_url: The url of the repository to use.
        :param str test_path: Set the test path to use for unittest discovery.
            If both this and the corresponding config file option are set, this
            value will be used.
        :param str top_dir: The top dir to use for unittest discovery. This
            takes precedence over the value in the config file. (if one is
            present in the config file)
        :param str group_regex: Set a group regex to use for grouping tests
            together in the stestr scheduler. If both this and the
            corresponding config file option are set this value will be used.
        :param str blacklist_file: Path to a blacklist file, this file contains
            a separate regex exclude on each newline.
        :param str whitelist_file: Path to a whitelist file, this file contains
            a separate regex on each newline.
        :param str black_regex: Test rejection regex. If a test cases name
            matches on re.search() operation, it will be removed from the final
            test list.
        :param list filters: A list of string regex filters to initially apply
            on the test list. Tests that match any of the regexes will be used.
            (assuming any other filtering specified also uses it)
        """

        self.config = config or '.stestr.conf'
        self.repo_type = repo_type or 'file'
        self.repo_url = repo_url
        self.test_path = test_path
        self.top_dir = top_dir
        self.group_regex = group_regex
        self.blacklist_file = blacklist_file
        self.whitelist_file = whitelist_file
        self.black_regex = black_regex
        self.filters = filters

    def discover_test_cases(self, **kwargs):
        """Iterate over test_ids for a project
        This method will print the test_ids for tests in a project. You can
        filter the output just like with the run command to see exactly what
        will be run.
        """
        from stestr import config_file

        params = dict(config=self.config, repo_type=self.repo_type,
                      repo_url=self.repo_url, test_path=self.test_path,
                      top_dir=self.top_dir, group_regex=self.group_regex,
                      blacklist_file=self.blacklist_file,
                      whitelist_file=self.whitelist_file,
                      black_regex=self.black_regex, filters=self.filters)
        if kwargs:
            params.update(kwargs)
        ids = None
        config = params.pop('config')
        conf = config_file.TestrConf(config)
        filters = params.pop('filters')
        blacklist_file = params.pop('blacklist_file')
        whitelist_file = params.pop('whitelist_file')
        black_regex = params.pop('black_regex')
        cmd = conf.get_run_command(
            regexes=filters, repo_type=params['repo_type'],
            repo_url=params['repo_url'], group_regex=params['group_regex'],
            blacklist_file=blacklist_file, whitelist_file=whitelist_file,
            black_regex=black_regex, test_path=params['test_path'],
            top_dir=params['top_dir'])
        not_filtered = filters is None and blacklist_file is None\
            and whitelist_file is None and black_regex is None

        try:
            cmd.setUp()
            # List tests if the fixture has not already needed to to filter.
            if not_filtered:
                ids = cmd.list_tests()
            else:
                ids = cmd.test_ids
        except SystemExit as ex:
            raise RuntimeError("Error discovering test cases IDs with "
                               f"parameters: {params}") from ex
        finally:
            cmd.cleanUp()

        return sorted(ids)


FINDER = TestCasesFinder()


def discover_test_cases(finder=FINDER, **kwargs):
    return finder.discover_test_cases(**kwargs)


class TestCaseTimeoutError(_exception.TobikoException):
    message = ("Test case '{testcase_id}' timed out after {timeout} seconds "
               "at:\n{stack}")


class TestCase(testtools.TestCase):

    _capture_log = False
    _capture_log_level = logging.DEBUG
    _capture_log_logger = logging.root
    _testcase_timeout: _time.Seconds = None

    @classmethod
    def setUpClass(cls):
        super(TestCase, cls).setUpClass()
        config = _config.tobiko_config()
        cls._capture_log = config.logging.capture_log
        cls._testcase_timeout = _time.to_seconds(cls._testcase_timeout or
                                                 config.testcase.timeout or
                                                 None)

    def setUp(self):
        super(TestCase, self).setUp()
        self._push_test_case()
        self._setup_capture_log()
        self._setup_testcase_timeout()

    def _setup_capture_log(self):
        if self._capture_log:
            self.useFixture(_logging.CaptureLogFixture(
                test_case_id=self.id(),
                level=self._capture_log_level,
                logger=self._capture_log_logger))

    def _setup_testcase_timeout(self):
        timeout = self._testcase_timeout
        if timeout is not None:
            self.useFixture(_itimer.itimer(
                delay=timeout,
                on_timeout=self._on_testcase_timeout))

    def _on_testcase_timeout(self, _signal_number, frame):
        stack = traceback.extract_stack(frame)
        for test_method_index, summary in enumerate(stack):
            if self._testMethodName == summary.name:
                stack = stack[test_method_index:]
                break

        formatted_stack = ''.join(traceback.format_list(stack))
        timeout = self._testcase_timeout
        raise TestCaseTimeoutError(testcase_id=self.id(), timeout=timeout,
                                   stack=formatted_stack)

    def _push_test_case(self):
        push_test_case(self)
        self.addCleanup(self._pop_test_case)

    def _pop_test_case(self):
        self.assertIs(self, pop_test_case())


class TestCasesManager(object):

    def __init__(self):
        self._test_cases: typing.List[TestCase] = []

    def get_test_case(self) -> TestCase:
        try:
            return self._test_cases[-1]
        except IndexError:
            return DUMMY_TEST_CASE

    def pop_test_case(self) -> TestCase:
        return self._test_cases.pop()

    def push_test_case(self, test_case: TestCase):
        _exception.check_valid_type(test_case, TestCase)
        self._test_cases.append(test_case)


TEST_CASES = TestCasesManager()


def push_test_case(test_case: testtools.TestCase,
                   manager: TestCasesManager = TEST_CASES):
    return manager.push_test_case(test_case=test_case)


def pop_test_case(manager: TestCasesManager = TEST_CASES) -> \
        testtools.TestCase:
    return manager.pop_test_case()


def get_test_case(manager: TestCasesManager = TEST_CASES) -> \
        testtools.TestCase:
    return manager.get_test_case()


class DummyTestCase(TestCase):

    def runTest(self):
        pass


DUMMY_TEST_CASE = DummyTestCase()


def run_test(test_case: testtools.TestCase,
             test_result: testtools.TestResult = None) -> testtools.TestResult:
    test_result = test_result or testtools.TestResult()
    test_case.run(test_result)
    return test_result


def assert_in(needle, haystack, message: typing.Optional[str] = None,
              manager: TestCasesManager = TEST_CASES):
    get_test_case(manager=manager).assertIn(needle, haystack, message)


def get_skipped_test_cases(test_result: testtools.TestResult,
                           skip_reason: typing.Optional[str] = None):
    if skip_reason is not None:
        assert_in(skip_reason, test_result.skip_reasons)
        return test_result.skip_reasons[skip_reason]
    else:
        skipped_test_cases = list()
        for cases in test_result.skip_reasons.values():
            skipped_test_cases.extend(cases)
        return skipped_test_cases


def assert_test_case_was_skipped(test_case: testtools.TestCase,
                                 test_result: testtools.TestResult,
                                 skip_reason: str = None,
                                 manager: TestCasesManager = TEST_CASES):
    skipped_tests = get_skipped_test_cases(test_result=test_result,
                                           skip_reason=skip_reason)
    assert_in(test_case, skipped_tests, manager=manager)
